from cifkit.utils import string_parser

from SAF.features.environment.util import _get_site_labels_per_element, count_first_second_min_dist, extract_best_labels


def compute_homoatomic_dist_by_site_shortest_dist(connections, R_best_label, M_best_label, X_best_label):
    """Compute the shortest homoatomic distances normalized by the
    shortest distance encountered at the best site for each element."""
    R = string_parser.get_atom_type_from_label(R_best_label)
    M = string_parser.get_atom_type_from_label(M_best_label)
    X = string_parser.get_atom_type_from_label(X_best_label)
    R_homoatomic_shortest_dist = float("inf")
    M_homoatomic_shortest_dist = float("inf")
    X_homoatomic_shortest_dist = float("inf")
    R_min_dist = float("inf")
    M_min_dist = float("inf")
    X_min_dist = float("inf")
    # Process each site
    for site_label, site_connections in connections.items():
        # Compute the shortest distance at this site
        current_min_dist = min([dist for _, dist, _, _ in site_connections])
        # Update the minimum distances for A and B if this is their best site
        if site_label == R_best_label:
            R_min_dist = current_min_dist
            R_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(R)])
        if site_label == M_best_label:
            M_min_dist = current_min_dist
            M_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(M)])
        if site_label == X_best_label:
            X_min_dist = current_min_dist
            X_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(X)])
    # Normalize the shortest distances by the shortest distances at the best sites
    R_homoatomic_dist_by_shortest_dist = R_homoatomic_shortest_dist / R_min_dist
    M_homoatomic_dist_by_shortest_dist = M_homoatomic_shortest_dist / M_min_dist
    X_homoatomic_dist_by_shortest_dist = X_homoatomic_shortest_dist / X_min_dist
    return (
        R_homoatomic_dist_by_shortest_dist,
        M_homoatomic_dist_by_shortest_dist,
        X_homoatomic_dist_by_shortest_dist,
    )


def compute_avg_homoatomic_dist_by_site_shortest_dist(connections, R, M, X):
    """Compute the average of the homoatomic distances normalized by the
    shortest distance encountered at each site for two distinct element
    types within a given set of connections."""
    # Shortest distance per site label
    element_to_site_labels = _get_site_labels_per_element(connections)
    R_total_homoatomic_dist_by_shortest_dist = 0.0
    M_total_homoatomic_dist_by_shortest_dist = 0.0
    X_total_homoatomic_dist_by_shortest_dist = 0.0
    for site_label, site_connections in connections.items():
        shortest_dist_per_site = site_connections[0][1]
        for connection in site_connections:
            other_label, dist, _, _ = connection
            if site_label.startswith(R) and other_label.startswith(R):
                R_total_homoatomic_dist_by_shortest_dist += dist / shortest_dist_per_site
                break
            if site_label.startswith(M) and other_label.startswith(M):
                M_total_homoatomic_dist_by_shortest_dist += dist / shortest_dist_per_site
                break
            if site_label.startswith(X) and other_label.startswith(X):
                X_total_homoatomic_dist_by_shortest_dist += dist / shortest_dist_per_site
                break
    R_avg_homoatomic_dist_by_shortest_dist = R_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[R])
    M_avg_homoatomic_dist_by_shortest_dist = M_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[M])
    X_avg_homoatomic_dist_by_shortest_dist = X_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[X])
    return (
        R_avg_homoatomic_dist_by_shortest_dist,
        M_avg_homoatomic_dist_by_shortest_dist,
        X_avg_homoatomic_dist_by_shortest_dist,
    )


def get_R_and_M_and_X_count_in_best_label_per_element(connections, R, M, X):
    """Compute the occurrences of element types R, M, X at the shortest
    distances within their respective best labeled site for each
    element.

    The best site label is determined with the label with the shortest
    distance pair.
    """
    R_count_at_R_shortest_dist = 0
    R_count_at_M_shortest_dist = 0
    R_count_at_X_shortest_dist = 0
    M_count_at_R_shortest_dist = 0
    M_count_at_M_shortest_dist = 0
    M_count_at_X_shortest_dist = 0
    X_count_at_R_shortest_dist = 0
    X_count_at_M_shortest_dist = 0
    X_count_at_X_shortest_dist = 0
    dist_count_per_label = count_first_second_min_dist(connections)
    best_label_dict = extract_best_labels(dist_count_per_label)
    best_R_site_label = best_label_dict[R]["best_label"]
    best_R_shortest_dist = best_label_dict[R]["best_label_details"]["shortest_dist"]
    best_M_site_label = best_label_dict[M]["best_label"]
    best_M_shortest_dist = best_label_dict[M]["best_label_details"]["shortest_dist"]
    best_X_site_label = best_label_dict[X]["best_label"]
    best_X_shortest_dist = best_label_dict[X]["best_label_details"]["shortest_dist"]
    for connection in connections[best_R_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_R_shortest_dist:
            if other_label.startswith(R):
                R_count_at_R_shortest_dist += 1
            if other_label.startswith(M):
                M_count_at_R_shortest_dist += 1
            if other_label.startswith(X):
                X_count_at_R_shortest_dist += 1
    for connection in connections[best_M_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_M_shortest_dist:
            if other_label.startswith(R):
                R_count_at_M_shortest_dist += 1
            if other_label.startswith(M):
                M_count_at_M_shortest_dist += 1
            if other_label.startswith(X):
                X_count_at_M_shortest_dist += 1
    for connection in connections[best_X_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_X_shortest_dist:
            if other_label.startswith(R):
                R_count_at_X_shortest_dist += 1
            if other_label.startswith(M):
                M_count_at_X_shortest_dist += 1
            if other_label.startswith(X):
                X_count_at_X_shortest_dist += 1
    return (
        R_count_at_R_shortest_dist,
        R_count_at_M_shortest_dist,
        R_count_at_X_shortest_dist,
        M_count_at_R_shortest_dist,
        M_count_at_M_shortest_dist,
        M_count_at_X_shortest_dist,
        X_count_at_R_shortest_dist,
        X_count_at_M_shortest_dist,
        X_count_at_X_shortest_dist,
    )


def get_avg_R_and_M_and_X_count_in_per_element(connections, R, M, X):
    """Compute the occurrences of element types R, M, and X at the
    shortest distances within their respective site for all site
    labels."""
    R_total_count_at_R_shortest_dist = 0
    R_total_count_at_M_shortest_dist = 0
    R_total_count_at_X_shortest_dist = 0
    M_total_count_at_R_shortest_dist = 0
    M_total_count_at_M_shortest_dist = 0
    M_total_count_at_X_shortest_dist = 0
    X_total_count_at_R_shortest_dist = 0
    X_total_count_at_M_shortest_dist = 0
    X_total_count_at_X_shortest_dist = 0
    dist_count_per_label = count_first_second_min_dist(connections)
    best_labels = extract_best_labels(dist_count_per_label)
    R_site_label_count = best_labels[R]["label_count"]
    M_site_label_count = best_labels[M]["label_count"]
    X_site_label_count = best_labels[X]["label_count"]

    for site_label, connection_data in connections.items():
        parsed_element = string_parser.get_atom_type_from_label(site_label)
        shortest_dist_per_site_label = dist_count_per_label[site_label]["shortest_dist"]
        for connection in connection_data:
            other_label, dist, _, _ = connection
            if parsed_element == R and dist == shortest_dist_per_site_label:
                if other_label.startswith(R):
                    R_total_count_at_R_shortest_dist += 1
                elif other_label.startswith(M):
                    R_total_count_at_M_shortest_dist += 1
                elif other_label.startswith(X):
                    R_total_count_at_X_shortest_dist += 1
            elif parsed_element == M and dist == shortest_dist_per_site_label:
                if other_label.startswith(R):
                    M_total_count_at_R_shortest_dist += 1
                elif other_label.startswith(M):
                    M_total_count_at_M_shortest_dist += 1
                elif other_label.startswith(X):
                    M_total_count_at_X_shortest_dist += 1
            elif parsed_element == X and dist == shortest_dist_per_site_label:
                if other_label.startswith(R):
                    X_total_count_at_R_shortest_dist += 1
                elif other_label.startswith(M):
                    X_total_count_at_M_shortest_dist += 1
                elif other_label.startswith(X):
                    X_total_count_at_X_shortest_dist += 1
    # Calculating average counts
    R_avg_count_at_R = R_total_count_at_R_shortest_dist / R_site_label_count
    R_avg_count_at_M = R_total_count_at_M_shortest_dist / M_site_label_count
    R_avg_count_at_X = R_total_count_at_X_shortest_dist / X_site_label_count
    M_avg_count_at_R = M_total_count_at_R_shortest_dist / R_site_label_count
    M_avg_count_at_M = M_total_count_at_M_shortest_dist / M_site_label_count
    M_avg_count_at_X = M_total_count_at_X_shortest_dist / X_site_label_count
    X_avg_count_at_R = X_total_count_at_R_shortest_dist / R_site_label_count
    X_avg_count_at_M = X_total_count_at_M_shortest_dist / M_site_label_count
    X_avg_count_at_X = X_total_count_at_X_shortest_dist / X_site_label_count

    return (
        R_avg_count_at_R,
        R_avg_count_at_M,
        R_avg_count_at_X,
        M_avg_count_at_R,
        M_avg_count_at_M,
        M_avg_count_at_X,
        X_avg_count_at_R,
        X_avg_count_at_M,
        X_avg_count_at_X,
    )



from SAF.features.wyc.binary import compute_features


def test_compute_features_ThSb(ThSb_cif, ThSb_elements_tuple):
    result, uni_result = compute_features(ThSb_cif, ThSb_elements_tuple)
    # Let A be Th and B be Sb
    expected = {
        "WYK_A_lowest_wyckoff": 1,
        "WYK_B_lowest_wyckoff": 1,
        "WYK_identical_lowest_wyckoff_count": 2,
        "WYK_A_sites_total": 1,
        "WYK_B_sites_total": 1,
        "WYK_A_multiplicity_total": 1,
        "WYK_B_multiplicity_total": 1,
    }
    uni_expected = {
        "UNI_WYK_lowest_wyckoff": 1,
    }
    assert result == expected
    assert uni_result == uni_expected


def test_compute_features_Th7Rh3(Th7Rh3_cif, ThRh_elements_tuple):
    result, uni_result = compute_features(Th7Rh3_cif, ThRh_elements_tuple)
    # Let A be Th and B be Rh
    expected = {
        "WYK_A_lowest_wyckoff": 2,
        "WYK_B_lowest_wyckoff": 6,
        "WYK_identical_lowest_wyckoff_count": 1,
        "WYK_A_sites_total": 3,
        "WYK_B_sites_total": 1,
        "WYK_A_multiplicity_total": 14,
        "WYK_B_multiplicity_total": 6,
    }
    uni_expected = {"UNI_WYK_lowest_wyckoff": 2}
    assert result == expected
    assert uni_result == uni_expected



from deepdiff import DeepDiff

from SAF.features.generator import compute_binary_features


def test_generate_binary_features():
    file_path = "tests/cif/binary/ThSb.cif"
    actual_features, actual_uni_features = compute_binary_features(file_path)
    expected_features = {
        "Entry": "ThSb",
        "Formula": "ThSb",
        "Structure": "CsCl",
        "A": "Th",
        "B": "Sb",
        "INT_AA_dist": 3.847,
        "INT_BB_dist": 3.847,
        "INT_AB_dist": 3.332,
        "INT_Asize": 1.798,
        "INT_Bsize": 1.434,
        "INT_Asize_by_Bsize": 1.2538354253835426,
        "INT_AA_dist_by_2_by_Asize": 1.0697997775305894,
        "INT_BB_dist_by_2_by_Bsize": 1.341352859135286,
        "INT_AB_dist_by_ABsizes": 1.0309405940594059,
        "INT_Asize_ref": 1.8591213139280958,
        "INT_Bsize_ref": 1.4728786860719039,
        "INT_percent_diff_A_by_100": 0.03399405668970841,
        "INT_percent_diff_B_by_100": 0.027112054443447647,
        "INT_dist_AA_minus_ref_diff": 0.033469553455629956,
        "INT_dist_BB_minus_ref_diff": 0.2342715435030393,
        "INT_dist_AB_minus_ref_diff": 0.0,
        "INT_R_factor": 0.0018906593863675784,
        "INT_UNI_shortest_homoatomic_dist": 3.847,
        "INT_UNI_shortest_heteroatomic_dist": 3.332,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 1.0697997775305894,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 1.0309405940594059,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": 1.0346285557535135,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": 1.0,
        "INT_UNI_highest_refined_percent_diff_abs": 0.03399405668970841,
        "INT_UNI_lowest_refined_percent_diff_abs": 0.027112054443447647,
        "INT_UNI_refined_packing_efficiency": 0.7078488680834484,
        "WYK_A_lowest_wyckoff": 1,
        "WYK_B_lowest_wyckoff": 1,
        "WYK_identical_lowest_wyckoff_count": 2,
        "WYK_A_sites_total": 1,
        "WYK_B_sites_total": 1,
        "WYK_A_multiplicity_total": 1,
        "WYK_B_multiplicity_total": 1,
        "ENV_A_shortest_dist_count": 8,
        "ENV_B_shortest_dist_count": 8,
        "ENV_A_avg_shortest_dist_count": 8.0,
        "ENV_B_avg_shortest_dist_count": 8.0,
        "ENV_A_shortest_tol_dist_count": 8,
        "ENV_B_shortest_tol_dist_count": 8,
        "ENV_A_avg_shortest_dist_within_tol_count": 8.0,
        "ENV_B_avg_shortest_dist_within_tol_count": 8.0,
        "ENV_A_second_by_first_shortest_dist": 1.154561824729892,
        "ENV_B_second_by_first_shortest_dist": 1.154561824729892,
        "ENV_A_avg_second_by_first_shortest_dist": 1.154561824729892,
        "ENV_B_avg_second_by_first_shortest_dist": 1.154561824729892,
        "ENV_A_second_shortest_dist_count": 6,
        "ENV_B_second_shortest_dist_count": 6,
        "ENV_A_avg_second_shortest_dist_count": 6.0,
        "ENV_B_avg_second_shortest_dist_count": 6.0,
        "ENV_A_homoatomic_dist_by_shortest_dist": 1.154561824729892,
        "ENV_B_homoatomic_dist_by_shortest_dist": 1.154561824729892,
        "ENV_A_avg_homoatomic_dist_by_shortest_dist": 1.154561824729892,
        "ENV_B_avg_homoatomic_dist_by_shortest_dist": 1.154561824729892,
        "ENV_A_count_at_A_shortest_dist": 0,
        "ENV_A_count_at_B_shortest_dist": 8,
        "ENV_A_avg_count_at_A_shortest_dist": 0.0,
        "ENV_A_avg_count_at_B_shortest_dist": 8.0,
        "ENV_B_count_at_A_shortest_dist": 8,
        "ENV_B_count_at_B_shortest_dist": 0,
        "ENV_B_avg_count_at_A_shortest_dist": 8.0,
        "ENV_B_avg_count_at_B_shortest_dist": 0.0,
        "CN_AVG_coordination_number": 14.0,
        "CN_AVG_A_atom_count": 7.0,
        "CN_AVG_B_atom_count": 7.0,
        "CN_AVG_polyhedron_volume": 113.86449999999999,
        "CN_AVG_central_atom_to_center_of_mass_dist": 0.0005,
        "CN_AVG_number_of_edges": 36.0,
        "CN_AVG_number_of_faces": 24.0,
        "CN_AVG_shortest_distance_to_face": 2.7945,
        "CN_AVG_shortest_distance_to_edge": 2.72,
        "CN_AVG_volume_of_inscribed_sphere": 91.398,
        "CN_AVG_packing_efficiency": 0.803,
        "CN_MIN_coordination_number": 14.0,
        "CN_MIN_A_atom_count": 7.0,
        "CN_MIN_B_atom_count": 7.0,
        "CN_MIN_polyhedron_volume": 113.86449999999999,
        "CN_MIN_central_atom_to_center_of_mass_dist": 0.0005,
        "CN_MIN_number_of_edges": 36.0,
        "CN_MIN_number_of_faces": 24.0,
        "CN_MIN_shortest_distance_to_face": 2.7945,
        "CN_MIN_shortest_distance_to_edge": 2.72,
        "CN_MIN_volume_of_inscribed_sphere": 91.398,
        "CN_MIN_packing_efficiency": 0.803,
        "CN_MAX_coordination_number": 14.0,
        "CN_MAX_A_atom_count": 7.0,
        "CN_MAX_B_atom_count": 7.0,
        "CN_MAX_polyhedron_volume": 113.86449999999999,
        "CN_MAX_central_atom_to_center_of_mass_dist": 0.0005,
        "CN_MAX_number_of_edges": 36.0,
        "CN_MAX_number_of_faces": 24.0,
        "CN_MAX_shortest_distance_to_face": 2.7945,
        "CN_MAX_shortest_distance_to_edge": 2.72,
        "CN_MAX_volume_of_inscribed_sphere": 91.398,
        "CN_MAX_packing_efficiency": 0.803,
    }
    # THis works but just unstable.
    diff = DeepDiff(expected_features, actual_features, significant_digits=3)
    assert diff == {}
    expected_uni_features = {
        "Entry": "ThSb",
        "Formula": "ThSb",
        "INT_UNI_shortest_homoatomic_dist": 3.847,
        "INT_UNI_shortest_heteroatomic_dist": 3.332,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 1.0697997775305894,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 1.0309405940594059,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": 1.0346285557535135,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": 1.0,
        "INT_UNI_highest_refined_percent_diff_abs": 0.03399405668970841,
        "INT_UNI_lowest_refined_percent_diff_abs": 0.027112054443447647,
        "INT_UNI_refined_packing_efficiency": 0.7078488680834484,
        "UNI_WYK_lowest_wyckoff": 1,
    }
    diff = DeepDiff(expected_uni_features, actual_uni_features, significant_digits=3)
    assert diff == {}



from SAF.utils.bond import get_min_distances_by_labels

binary_sample_data = {
    ("Th", "Th"): 1.1,
    ("Os", "Os"): 1.2,
    ("Th", "Os"): 1.3,
}


def test_get_shortest_bond_distances_by_AB():
    result = get_min_distances_by_labels(binary_sample_data, ["Th", "Os"], labels=["A", "B"])
    expected = {
        "AA": 1.1,
        "BB": 1.2,
        "AB": 1.3,
    }
    assert result == expected


ternary_sample_data = {
    ("U", "U"): 1.0,
    ("U", "Rh"): 1.6,
    ("U", "In"): 1.7,
    ("Rh", "Rh"): 1.4,
    ("Rh", "In"): 1.8,
    ("In", "In"): 1.5,
}


def test_get_shortest_bond_distances_by_RMX():
    result = get_min_distances_by_labels(ternary_sample_data, ["U", "Rh", "In"], labels=["R", "M", "X"])
    expected = {
        "RR": 1.0,
        "RM": 1.6,
        "RX": 1.7,
        "MM": 1.4,
        "MX": 1.8,
        "XX": 1.5,
    }
    assert result == expected


quaternary_sample_data = {
    ("Cu", "Cu"): 1.0,
    ("U", "U"): 1.1,
    ("Rh", "Rh"): 1.2,
    ("In", "In"): 1.3,
    ("Cu", "U"): 1.4,
    ("Cu", "Rh"): 1.5,
    ("Cu", "In"): 1.6,
    ("U", "Rh"): 1.7,
    ("U", "In"): 1.8,
    ("Rh", "In"): 1.9,
}


def test_get_shortest_bond_distances_by_quaternary():
    result = get_min_distances_by_labels(quaternary_sample_data, ["Cu", "U", "Rh", "In"], labels=["A", "B", "C", "D"])
    expected = {
        "AA": 1.0,
        "BB": 1.1,
        "CC": 1.2,
        "DD": 1.3,
        "AB": 1.4,
        "AC": 1.5,
        "AD": 1.6,
        "BC": 1.7,
        "BD": 1.8,
        "CD": 1.9,
    }
    assert result == expected



import pytest

from SAF.features.coordination.quaternary import compute_features


def test_features(Tb4RhInGe4_cif, TbRhInGe_elements_tuple):
    result = compute_features(Tb4RhInGe4_cif, TbRhInGe_elements_tuple)
    expected = {
        "CN_AVG_coordination_number": 13.068181818181818,
        "CN_AVG_A_atom_count": 6.9772727272727275,
        "CN_AVG_B_atom_count": 1.0909090909090908,
        "CN_AVG_C_atom_count": 1.2727272727272727,
        "CN_AVG_D_atom_count": 3.727272727272727,
        "CN_AVG_polyhedron_volume": 85.20697727272726,
        "CN_AVG_central_atom_to_center_of_mass_dist": 0.09595454545454546,
        "CN_AVG_number_of_edges": 33.20454545454545,
        "CN_AVG_number_of_faces": 22.136363636363637,
        "CN_AVG_shortest_distance_to_face": 2.3147045454545454,
        "CN_AVG_shortest_distance_to_edge": 2.3177499999999998,
        "CN_AVG_volume_of_inscribed_sphere": 55.812318181818185,
        "CN_AVG_packing_efficiency": 0.6347499999999999,
        "CN_MIN_coordination_number": 12.818181818181818,
        "CN_MIN_A_atom_count": 6.909090909090909,
        "CN_MIN_B_atom_count": 1.0909090909090908,
        "CN_MIN_C_atom_count": 1.0909090909090908,
        "CN_MIN_D_atom_count": 3.727272727272727,
        "CN_MIN_polyhedron_volume": 82.86009090909091,
        "CN_MIN_central_atom_to_center_of_mass_dist": 0.09081818181818181,
        "CN_MIN_number_of_edges": 32.45454545454545,
        "CN_MIN_number_of_faces": 21.636363636363637,
        "CN_MIN_shortest_distance_to_face": 2.2578181818181817,
        "CN_MIN_shortest_distance_to_edge": 2.225818181818182,
        "CN_MIN_volume_of_inscribed_sphere": 53.12227272727273,
        "CN_MIN_packing_efficiency": 0.6080909090909091,
        "CN_MAX_coordination_number": 13.272727272727273,
        "CN_MAX_A_atom_count": 7.0,
        "CN_MAX_B_atom_count": 1.0909090909090908,
        "CN_MAX_C_atom_count": 1.4545454545454546,
        "CN_MAX_D_atom_count": 3.727272727272727,
        "CN_MAX_polyhedron_volume": 87.19936363636363,
        "CN_MAX_central_atom_to_center_of_mass_dist": 0.11136363636363637,
        "CN_MAX_number_of_edges": 33.81818181818182,
        "CN_MAX_number_of_faces": 22.545454545454547,
        "CN_MAX_shortest_distance_to_face": 2.344727272727272,
        "CN_MAX_shortest_distance_to_edge": 2.3777272727272725,
        "CN_MAX_volume_of_inscribed_sphere": 57.506181818181815,
        "CN_MAX_packing_efficiency": 0.644,
    }
    for key in result:
        assert result[key] == pytest.approx(expected[key], abs=0.01)



from SAF.features.wyc.ternary import compute_features


def test_compute_features_ThSb(URhIn_cif, URhIn_elements_tuple):
    # Let R = U, M = Rh, X = In
    wyk_actual, wyk_uni_actual = compute_features(URhIn_cif, URhIn_elements_tuple)
    expected = {
        "WYK_R_lowest_wyckoff": 3,
        "WYK_M_lowest_wyckoff": 1,
        "WYK_X_lowest_wyckoff": 3,
        "WYK_identical_lowest_wyckoff_count": 1,
        "WYK_R_sites_total": 1,
        "WYK_M_sites_total": 2,
        "WYK_X_sites_total": 1,
        "WYK_R_multiplicity_total": 3,
        "WYK_M_multiplicity_total": 3,
        "WYK_X_multiplicity_total": 3,
    }
    uni_expected = {"UNI_WYK_lowest_wyckoff": 1}
    assert wyk_actual == expected
    assert wyk_uni_actual == uni_expected



import pytest
from cifkit import Cif

"""
Binary
"""


@pytest.fixture
def ThSb_cif():
    cif = Cif("tests/cif/binary/ThSb.cif", supercell_size=2)
    cif.compute_connections()
    return cif


@pytest.fixture
def Th7Rh3_cif():
    cif = Cif("tests/cif/binary/Th7Rh3.cif", supercell_size=2)
    cif.compute_connections()
    cif.compute_CN()
    return cif


@pytest.fixture
def Dy2Co17_cif():
    cif = Cif("tests/cif/binary/Dy2Co17.cif", supercell_size=2)
    cif.compute_connections()
    return cif


@pytest.fixture
def RhSb2_cif():
    cif = Cif("tests/cif/binary/RhSb2.cif", supercell_size=2)
    cif.compute_connections()
    cif.compute_CN()
    return cif


@pytest.fixture
def RhSb_elements_tuple():
    return ("Rh", "Sb")


@pytest.fixture
def ThRh_elements_tuple():
    return ("Th", "Rh")


@pytest.fixture
def ThSb_elements_tuple():
    return ("Th", "Sb")


"""
Ternary
"""


@pytest.fixture
def URhIn_cif():
    cif = Cif("tests/cif/ternary/URhIn.cif", supercell_size=2)
    cif.compute_connections()
    cif.compute_CN()
    return cif


@pytest.fixture
def URhIn_elements_tuple():
    return ("U", "Rh", "In")


"""
Quaternary
"""


@pytest.fixture
def Tb4RhInGe4_cif():
    # Tested that supercell size 2 or 3 both work
    cif = Cif("tests/cif/quaternary/Tb4RhInGe4.cif", supercell_size=2)
    cif.compute_connections()
    cif.compute_CN()
    return cif


@pytest.fixture
def TbRhInGe_elements_tuple():
    return ("Tb", "Rh", "In", "Ge")



from cifkit import Cif
from cifkit.data import radius_optimization as radius_opt

from SAF.features.interatomic import helper
from SAF.utils import bond, packing


def compute_features(cif: Cif, use_size_constraint: bool, elements):
    A, B = elements
    A_CIF_rad = cif.radius_values[A]["CIF_radius"]
    B_CIF_rad = cif.radius_values[B]["CIF_radius"]
    CIF_rad_refined, obj_value = radius_opt.get_refined_CIF_radius(
        [A, B], cif.shortest_bond_pair_distance, elements_ordered=False, use_size_constraint=use_size_constraint
    )
    min_bond_dists = bond.get_min_distances_by_labels(cif.shortest_bond_pair_distance, [A, B], labels=["A", "B"])
    AA_dist = min_bond_dists["AA"]
    BB_dist = min_bond_dists["BB"]
    AB_dist = min_bond_dists["AB"]
    A_CIF_rad_refined = float(CIF_rad_refined[A])
    B_CIF_rad_refined = float(CIF_rad_refined[B])
    # Distances
    AA_dist_by_2 = AA_dist / 2
    BB_dist_by_2 = BB_dist / 2
    # Refined distances
    AA_ref_radius_sum = 2 * A_CIF_rad_refined
    BB_ref_radius_sum = 2 * B_CIF_rad_refined
    AB_ref_radius_sum = A_CIF_rad_refined + B_CIF_rad_refined
    # Percent differences between refined and CIF radii
    percent_diff_A = (A_CIF_rad_refined - A_CIF_rad) / A_CIF_rad
    percent_diff_B = (B_CIF_rad_refined - B_CIF_rad) / B_CIF_rad
    # Interatomic distance features
    Asize_by_Bsize = A_CIF_rad / B_CIF_rad
    AA_dist_by_2_by_Asize = AA_dist_by_2 / A_CIF_rad
    BB_dist_by_2_by_Bsize = BB_dist_by_2 / B_CIF_rad
    AB_dist_by_ABsizes = AB_dist / (A_CIF_rad + B_CIF_rad)
    AA_dist_minus_ref_diff = (AA_dist - AA_ref_radius_sum) / AA_dist
    BB_dist_minus_ref_diff = (BB_dist - BB_ref_radius_sum) / BB_dist
    AB_dist_minus_ref_diff = (AB_dist - AB_ref_radius_sum) / AB_dist
    # UNI Homoatomic and heteroatomic distances
    homoatomic_dists = helper.get_bond_distances_by_labels(min_bond_dists, ["AA", "BB"])
    heteroatomic_dists = helper.get_bond_distances_by_labels(min_bond_dists, ["AB"])
    shortest_heteroatomic_dist = helper.get_min_from_dists(heteroatomic_dists)
    shortest_homoatomic_dist = helper.get_min_from_dists(homoatomic_dists)
    # Define the CIF and refined radii
    rad_dict = {
        "AA": A_CIF_rad,
        "BB": B_CIF_rad,
        "AB": (A_CIF_rad + B_CIF_rad),
    }
    refined_rad_dict = {
        "AA": A_CIF_rad_refined,
        "BB": B_CIF_rad_refined,
        "AB": (A_CIF_rad_refined + B_CIF_rad_refined),
    }
    shortest_homo_key = helper.get_shortest_homo_key(min_bond_dists, shortest_homoatomic_dist)
    shortest_hetero_key = helper.get_shortest_hetero_key(min_bond_dists, shortest_heteroatomic_dist)
    uni_features = helper.get_shortest_homo_heteroatomic_features(
        shortest_homoatomic_dist,
        shortest_homo_key,
        shortest_heteroatomic_dist,
        shortest_hetero_key,
        rad_dict,
        refined_rad_dict,
    )
    percent_diffs = [percent_diff_A, percent_diff_B]
    highest_refined_percent_diff = helper.get_highest_refined_percent_diff(percent_diffs)
    lowest_refined_percent_diff = helper.get_lowest_refined_percent_diff(percent_diffs)
    # Packing efficiency
    packing_efficiency = packing.compute_efficiency(
        cif,
        {
            A: A_CIF_rad_refined,
            B: B_CIF_rad_refined,
        },
    )
    data = {
        "Entry": cif.file_name_without_ext,
        "Formula": cif.formula,
        "Structure": cif.structure,
        "A": A,
        "B": B,
        "INT_AA_dist": AA_dist,
        "INT_BB_dist": BB_dist,
        "INT_AB_dist": AB_dist,
        "INT_Asize": A_CIF_rad,
        "INT_Bsize": B_CIF_rad,
        "INT_Asize_by_Bsize": Asize_by_Bsize,
        "INT_AA_dist_by_2_by_Asize": AA_dist_by_2_by_Asize,
        "INT_BB_dist_by_2_by_Bsize": BB_dist_by_2_by_Bsize,
        "INT_AB_dist_by_ABsizes": AB_dist_by_ABsizes,
        "INT_Asize_ref": A_CIF_rad_refined,
        "INT_Bsize_ref": B_CIF_rad_refined,
        "INT_percent_diff_A_by_100": percent_diff_A,
        "INT_percent_diff_B_by_100": percent_diff_B,
        "INT_dist_AA_minus_ref_diff": AA_dist_minus_ref_diff,
        "INT_dist_BB_minus_ref_diff": BB_dist_minus_ref_diff,
        "INT_dist_AB_minus_ref_diff": AB_dist_minus_ref_diff,
        "INT_R_factor": obj_value,
        "INT_UNI_shortest_homoatomic_dist": shortest_homoatomic_dist,
        "INT_UNI_shortest_heteroatomic_dist": shortest_heteroatomic_dist,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_atom_size"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": uni_features["shortest_heteroatomic_dist_by_sum_of_atom_sizes"],
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_refined_atom_sizes"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": uni_features["shortest_heteroatomic_dist_by_refined_atom_sizes"],
        "INT_UNI_highest_refined_percent_diff_abs": highest_refined_percent_diff,
        "INT_UNI_lowest_refined_percent_diff_abs": lowest_refined_percent_diff,
        "INT_UNI_refined_packing_efficiency": packing_efficiency,
    }
    uni_data = {
        "Entry": cif.file_name_without_ext,
        "Formula": cif.formula,
        "INT_UNI_shortest_homoatomic_dist": shortest_homoatomic_dist,
        "INT_UNI_shortest_heteroatomic_dist": shortest_heteroatomic_dist,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_atom_size"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": uni_features["shortest_heteroatomic_dist_by_sum_of_atom_sizes"],
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_refined_atom_sizes"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": uni_features["shortest_heteroatomic_dist_by_refined_atom_sizes"],
        "INT_UNI_highest_refined_percent_diff_abs": highest_refined_percent_diff,
        "INT_UNI_lowest_refined_percent_diff_abs": lowest_refined_percent_diff,
        "INT_UNI_refined_packing_efficiency": packing_efficiency,
    }
    return data, uni_data



from deepdiff import DeepDiff

from SAF.features.generator import compute_quaternary_features


def test_generate_features():
    file_path = "tests/cif/quaternary/Tb4RhInGe4.cif"
    actual_features, actual_uni_features = compute_quaternary_features(file_path)
    expected_features = {
        "Entry": "Tb4RhInGe4",
        "Formula": "Tb4RhInGe4",
        "Structure": "Tb4RhInGe4",
        "A": "Tb",
        "B": "Rh",
        "C": "In",
        "D": "Ge",
        "INT_AA_dist": 3.538,
        "INT_BB_dist": 3.148,
        "INT_CC_dist": 4.264,
        "INT_DD_dist": 2.596,
        "INT_AB_dist": 3.076,
        "INT_AC_dist": 3.367,
        "INT_AD_dist": 2.917,
        "INT_BC_dist": 4.711,
        "INT_BD_dist": 2.47,
        "INT_CD_dist": 2.818,
        "INT_Asize": 1.764,
        "INT_Bsize": 1.345,
        "INT_Csize": 1.624,
        "INT_Dsize": 1.225,
        "INT_Asize_by_Bsize": 1.3115241635687733,
        "INT_Bsize_by_Csize": 0.8282019704433496,
        "INT_Csize_by_Dsize": 1.3257142857142856,
        "INT_Asize_by_Csize": 1.086206896551724,
        "INT_Asize_by_Dsize": 1.44,
        "INT_Bsize_by_Dsize": 1.0979591836734692,
        "INT_AA_dist_by_2_by_Asize": 1.0028344671201814,
        "INT_BB_dist_by_2_by_Bsize": 1.1702602230483272,
        "INT_CC_dist_by_2_by_Csize": 1.312807881773399,
        "INT_DD_dist_by_2_by_Dsize": 1.059591836734694,
        "INT_AB_dist_by_ABsizes": 0.9893856545513027,
        "INT_AC_dist_by_ACsizes": 0.9938016528925621,
        "INT_AD_dist_by_ADsizes": 0.9759116761458682,
        "INT_BC_dist_by_BCsizes": 1.5867295385651734,
        "INT_BD_dist_by_BDsizes": 0.961089494163424,
        "INT_CD_dist_by_CDsizes": 0.9891189891189891,
        "INT_Asize_ref": 1.6862073857087099,
        "INT_Bsize_ref": 1.4201916565448087,
        "INT_Csize_ref": 1.6861073857088038,
        "INT_Dsize_ref": 1.1604492903475334,
        "INT_percent_diff_A_by_100": -0.04410012148032321,
        "INT_percent_diff_B_by_100": 0.05590457735673513,
        "INT_percent_diff_C_by_100": 0.038243464106406215,
        "INT_percent_diff_D_by_100": -0.05269445685915647,
        "INT_AA_minus_ref_diff": 0.04680193006856418,
        "INT_BB_minus_ref_diff": 0.0977181343425612,
        "INT_CC_minus_ref_diff": 0.20914287724727781,
        "INT_DD_minus_ref_diff": 0.10597127091869539,
        "INT_AB_minus_ref_diff": -0.009882653528452056,
        "INT_AC_minus_ref_diff": -0.0015784886894902542,
        "INT_AD_minus_ref_diff": 0.02411495507156549,
        "INT_BC_minus_ref_diff": 0.3406285200056013,
        "INT_BD_minus_ref_diff": -0.04479390562442993,
        "INT_CD_minus_ref_diff": -0.010133667869530575,
        "INT_R_factor": 0.00930941081455393,
        "INT_UNI_shortest_homoatomic_dist": 2.596,
        "INT_UNI_shortest_heteroatomic_dist": 2.47,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 1.059591836734694,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 0.961089494163424,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": 1.1185322881375306,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": 0.9571265630634986,
        "INT_UNI_highest_refined_percent_diff_abs": 0.05590457735673513,
        "INT_UNI_lowest_refined_percent_diff_abs": 0.038243464106406215,
        "INT_UNI_refined_packing_efficiency": 0.6486230122109067,
        "WYK_A_lowest_wyckoff": 4,
        "WYK_B_lowest_wyckoff": 4,
        "WYK_C_lowest_wyckoff": 2,
        "WYK_D_lowest_wyckoff": 4,
        "WYK_identical_lowest_wyckoff_count": 1,
        "WYK_A_sites_total": 4,
        "WYK_B_sites_total": 1,
        "WYK_C_sites_total": 2,
        "WYK_D_sites_total": 4,
        "WYK_A_multiplicity_total": 16,
        "WYK_B_multiplicity_total": 4,
        "WYK_C_multiplicity_total": 4,
        "WYK_D_multiplicity_total": 16,
        "ENV_A_shortest_dist_count": 1,
        "ENV_B_shortest_dist_count": 2,
        "ENV_C_shortest_dist_count": 2,
        "ENV_D_shortest_dist_count": 2,
        "ENV_A_avg_shortest_dist_count": 1.75,
        "ENV_B_avg_shortest_dist_count": 2.0,
        "ENV_C_avg_shortest_dist_count": 2.0,
        "ENV_D_avg_shortest_dist_count": 1.25,
        "ENV_A_shortest_tol_dist_count": 6,
        "ENV_B_shortest_tol_dist_count": 4,
        "ENV_C_shortest_tol_dist_count": 4,
        "ENV_D_shortest_tol_dist_count": 3,
        "ENV_A_avg_shortest_dist_within_tol_count": 5.0,
        "ENV_B_avg_shortest_dist_within_tol_count": 4.0,
        "ENV_C_avg_shortest_dist_within_tol_count": 3.0,
        "ENV_D_avg_shortest_dist_within_tol_count": 1.75,
        "ENV_A_second_by_first_shortest_dist": 1.0205690778196779,
        "ENV_B_second_by_first_shortest_dist": 1.0291497975708501,
        "ENV_C_second_by_first_shortest_dist": 1.0184528034066713,
        "ENV_D_second_by_first_shortest_dist": 1.0291497975708501,
        "ENV_A_avg_second_by_first_shortest_dist": 1.0364821721769375,
        "ENV_B_avg_second_by_first_shortest_dist": 1.0291497975708501,
        "ENV_C_avg_second_by_first_shortest_dist": 1.0440696072782485,
        "ENV_D_avg_second_by_first_shortest_dist": 1.0680669343726787,
        "ENV_A_second_shortest_dist_count": 2,
        "EnV_B_second_shortest_dist_count": 1,
        "ENV_C_second_shortest_dist_count": 2,
        "ENV_D_second_shortest_dist_count": 1,
        "ENV_A_avg_second_shortest_dist_count": 2.0,
        "ENV_B_avg_second_shortest_dist_count": 1.0,
        "ENV_C_avg_second_shortest_dist_count": 2.0,
        "ENV_D_avg_second_shortest_dist_count": 1.25,
        "ENV_A_homoatomic_dist_by_shortest_dist": 1.2128899554336647,
        "ENV_B_homoatomic_dist_by_shortest_dist": 1.274493927125506,
        "ENV_C_homoatomic_dist_by_shortest_dist": 1.5131298793470547,
        "ENV_D_homoatomic_dist_by_shortest_dist": 1.0510121457489878,
        "ENV_A_avg_homoatomic_dist_by_shortest_dist": 1.215032553686275,
        "ENV_B_avg_homoatomic_dist_by_shortest_dist": 1.274493927125506,
        "ENV_C_avg_homoatomic_dist_by_shortest_dist": 1.4994220825306703,
        "ENV_D_avg_homoatomic_dist_by_shortest_dist": 1.0193165160610715,
        "ENV_A_count_at_A_shortest_dist": 0,
        "ENV_B_count_at_A_shortest_dist": 0,
        "ENV_C_count_at_A_shortest_dist": 0,
        "ENV_D_count_at_A_shortest_dist": 1,
        "ENV_A_avg_count_at_A_shortest_dist": 0.0,
        "ENV_B_avg_count_at_A_shortest_dist": 0.0,
        "ENV_C_avg_count_at_A_shortest_dist": 0.0,
        "ENV_D_avg_count_at_A_shortest_dist": 1.75,
        "ENV_A_count_at_B_shortest_dist": 0,
        "ENV_B_count_at_B_shortest_dist": 0,
        "ENV_C_count_at_B_shortest_dist": 0,
        "ENV_D_count_at_B_shortest_dist": 2,
        "ENV_A_avg_count_at_B_shortest_dist": 0.0,
        "ENV_B_avg_count_at_B_shortest_dist": 0.0,
        "ENV_C_avg_count_at_B_shortest_dist": 0.0,
        "ENV_D_avg_count_at_B_shortest_dist": 2.0,
        "ENV_A_count_at_C_shortest_dist": 0,
        "ENV_B_count_at_C_shortest_dist": 0,
        "ENV_C_count_at_C_shortest_dist": 0,
        "ENV_D_count_at_C_shortest_dist": 2,
        "ENV_A_avg_count_at_C_shortest_dist": 0.0,
        "ENV_B_avg_count_at_C_shortest_dist": 0.0,
        "ENV_C_avg_count_at_C_shortest_dist": 0.0,
        "ENV_D_avg_count_at_C_shortest_dist": 2.0,
        "ENV_A_count_at_D_shortest_dist": 0,
        "ENV_B_count_at_D_shortest_dist": 2,
        "ENV_C_count_at_D_shortest_dist": 0,
        "ENV_D_count_at_D_shortest_dist": 0,
        "ENV_A_avg_count_at_D_shortest_dist": 0.0,
        "ENV_B_avg_count_at_D_shortest_dist": 0.75,
        "ENV_C_avg_count_at_D_shortest_dist": 0.0,
        "ENV_D_avg_count_at_D_shortest_dist": 0.5,
        "CN_AVG_coordination_number": 13.068181818181818,
        "CN_AVG_A_atom_count": 6.9772727272727275,
        "CN_AVG_B_atom_count": 1.0909090909090908,
        "CN_AVG_C_atom_count": 1.2727272727272727,
        "CN_AVG_D_atom_count": 3.727272727272727,
        "CN_AVG_polyhedron_volume": 85.2066590909091,
        "CN_AVG_central_atom_to_center_of_mass_dist": 0.09586363636363636,
        "CN_AVG_number_of_edges": 33.20454545454545,
        "CN_AVG_number_of_faces": 22.136363636363637,
        "CN_AVG_shortest_distance_to_face": 2.3237954545454547,
        "CN_AVG_shortest_distance_to_edge": 2.3122045454545455,
        "CN_AVG_volume_of_inscribed_sphere": 56.64554545454546,
        "CN_AVG_packing_efficiency": 0.6419318181818182,
        "CN_MIN_coordination_number": 12.818181818181818,
        "CN_MIN_A_atom_count": 6.909090909090909,
        "CN_MIN_B_atom_count": 1.0909090909090908,
        "CN_MIN_C_atom_count": 1.0909090909090908,
        "CN_MIN_D_atom_count": 3.727272727272727,
        "CN_MIN_polyhedron_volume": 82.8599090909091,
        "CN_MIN_central_atom_to_center_of_mass_dist": 0.09072727272727273,
        "CN_MIN_number_of_edges": 32.45454545454545,
        "CN_MIN_number_of_faces": 21.636363636363637,
        "CN_MIN_shortest_distance_to_face": 2.266909090909091,
        "CN_MIN_shortest_distance_to_edge": 2.220272727272728,
        "CN_MIN_volume_of_inscribed_sphere": 53.95327272727274,
        "CN_MIN_packing_efficiency": 0.6151818181818182,
        "CN_MAX_coordination_number": 13.272727272727273,
        "CN_MAX_A_atom_count": 7.0,
        "CN_MAX_B_atom_count": 1.0909090909090908,
        "CN_MAX_C_atom_count": 1.4545454545454546,
        "CN_MAX_D_atom_count": 3.727272727272727,
        "CN_MAX_polyhedron_volume": 87.19890909090908,
        "CN_MAX_central_atom_to_center_of_mass_dist": 0.11127272727272727,
        "CN_MAX_number_of_edges": 33.81818181818182,
        "CN_MAX_number_of_faces": 22.545454545454547,
        "CN_MAX_shortest_distance_to_face": 2.3538181818181814,
        "CN_MAX_shortest_distance_to_edge": 2.3721818181818186,
        "CN_MAX_volume_of_inscribed_sphere": 58.341636363636354,
        "CN_MAX_packing_efficiency": 0.6512727272727273,
    }
    # FIXME: This tolernace is too high because the polyhedron volume constantly changes
    diff = DeepDiff(expected_features, actual_features, math_epsilon=1)
    assert diff == {}
    expected_uni_features = {
        "Entry": "Tb4RhInGe4",
        "Formula": "Tb4RhInGe4",
        "INT_UNI_shortest_homoatomic_dist": 2.596,
        "INT_UNI_shortest_heteroatomic_dist": 2.47,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 1.059591836734694,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 0.961089494163424,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": 1.1185322881375306,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": 0.9571265630634986,
        "INT_UNI_highest_refined_percent_diff_abs": 0.05590457735673513,
        "INT_UNI_lowest_refined_percent_diff_abs": 0.038243464106406215,
        "INT_UNI_refined_packing_efficiency": 0.6486230122109067,
        "UNI_WYK_lowest_wyckoff": 2,
    }
    diff = DeepDiff(expected_uni_features, actual_uni_features, math_epsilon=1)
    assert diff == {}



def get_bond_distances_by_labels(min_dists: dict[str, float], bond_labels: list[str]) -> dict[str, float]:
    """Extract bond distances for specified labels from a dictionary of
    minimum distances."""
    return {key: min_dists[key] for key in bond_labels}


def get_min_from_dists(dists: dict[str, float]) -> float:
    """Get the minimum value from a dictionary of distances."""
    return min(dists.values())


def get_shortest_homo_key(min_bond_dists, shortest_homoatomic_dist):
    return [k for k, v in min_bond_dists.items() if v == shortest_homoatomic_dist][0]


def get_shortest_hetero_key(min_bond_dists, shortest_heteroatomic_dist):
    return [k for k, v in min_bond_dists.items() if v == shortest_heteroatomic_dist][0]


def get_highest_refined_percent_diff(percent_diffs):
    return max([abs(p) for p in percent_diffs])


def get_lowest_refined_percent_diff(percent_diffs):
    return min([abs(p) for p in percent_diffs])


def get_shortest_homo_heteroatomic_features(
    shortest_homoatomic_dist,
    shortest_homo_key,
    shortest_heteroatomic_dist,
    shortest_hetero_key,
    rad_dict,
    refined_rad_dict,
):
    return {
        "shortest_homoatomic_dist_by_2_by_atom_size": (shortest_homoatomic_dist / 2) / rad_dict[shortest_homo_key],
        "shortest_homoatomic_dist_by_2_by_refined_atom_sizes": (shortest_homoatomic_dist / 2) / refined_rad_dict[shortest_homo_key],
        "shortest_heteroatomic_dist_by_sum_of_atom_sizes": shortest_heteroatomic_dist / rad_dict[shortest_hetero_key],
        "shortest_heteroatomic_dist_by_refined_atom_sizes": shortest_heteroatomic_dist / refined_rad_dict[shortest_hetero_key],
    }



from cifkit.utils import string_parser

from SAF.features.environment.util import _get_site_labels_per_element, count_first_second_min_dist, extract_best_labels


def compute_homoatomic_dist_by_site_shortest_dist(connections, A_best_label, B_best_label, C_best_label, D_best_label):
    """Compute the shortest homoatomic distances normalized by the
    shortest distance encountered at the best site for each element."""
    A = string_parser.get_atom_type_from_label(A_best_label)
    B = string_parser.get_atom_type_from_label(B_best_label)
    C = string_parser.get_atom_type_from_label(C_best_label)
    D = string_parser.get_atom_type_from_label(D_best_label)
    A_homoatomic_shortest_dist = float("inf")
    B_homoatomic_shortest_dist = float("inf")
    C_homoatomic_shortest_dist = float("inf")
    D_homoatomic_shortest_dist = float("inf")
    A_min_dist = float("inf")
    B_min_dist = float("inf")
    C_min_dist = float("inf")
    D_min_dist = float("inf")
    # Process each site
    for site_label, site_connections in connections.items():
        # Compute the shortest distance at this site
        current_min_dist = min([dist for _, dist, _, _ in site_connections])
        # Update the minimum distances for A, B, C, and D if this is their best site
        if site_label == A_best_label:
            A_min_dist = current_min_dist
            A_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(A)])
        if site_label == B_best_label:
            B_min_dist = current_min_dist
            B_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(B)])
        if site_label == C_best_label:
            C_min_dist = current_min_dist
            C_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(C)])
        if site_label == D_best_label:
            D_min_dist = current_min_dist
            D_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(D)])
    # Normalize the shortest distances by the shortest distances at the best sites
    A_homoatomic_dist_by_shortest_dist = A_homoatomic_shortest_dist / A_min_dist
    B_homoatomic_dist_by_shortest_dist = B_homoatomic_shortest_dist / B_min_dist
    C_homoatomic_dist_by_shortest_dist = C_homoatomic_shortest_dist / C_min_dist
    D_homoatomic_dist_by_shortest_dist = D_homoatomic_shortest_dist / D_min_dist
    return (
        A_homoatomic_dist_by_shortest_dist,
        B_homoatomic_dist_by_shortest_dist,
        C_homoatomic_dist_by_shortest_dist,
        D_homoatomic_dist_by_shortest_dist,
    )


def compute_avg_homoatomic_dist_by_site_shortest_dist(connections, A, B, C, D):
    """Compute the average of the homoatomic distances normalized by the
    shortest distance encountered at each site for four distinct element
    types within a given set of connections."""
    element_to_site_labels = _get_site_labels_per_element(connections)
    A_total_homoatomic_dist_by_shortest_dist = 0.0
    B_total_homoatomic_dist_by_shortest_dist = 0.0
    C_total_homoatomic_dist_by_shortest_dist = 0.0
    D_total_homoatomic_dist_by_shortest_dist = 0.0
    for site_label, site_connections in connections.items():
        shortest_dist_per_site = site_connections[0][1]
        for connection in site_connections:
            other_label, dist, _, _ = connection
            if site_label.startswith(A) and other_label.startswith(A):
                A_total_homoatomic_dist_by_shortest_dist += dist / shortest_dist_per_site
                break
            if site_label.startswith(B) and other_label.startswith(B):
                B_total_homoatomic_dist_by_shortest_dist += dist / shortest_dist_per_site
                break
            if site_label.startswith(C) and other_label.startswith(C):
                C_total_homoatomic_dist_by_shortest_dist += dist / shortest_dist_per_site
                break
            if site_label.startswith(D) and other_label.startswith(D):
                D_total_homoatomic_dist_by_shortest_dist += dist / shortest_dist_per_site
                break
    A_avg_homoatomic_dist_by_shortest_dist = A_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[A])
    B_avg_homoatomic_dist_by_shortest_dist = B_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[B])
    C_avg_homoatomic_dist_by_shortest_dist = C_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[C])
    D_avg_homoatomic_dist_by_shortest_dist = D_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[D])
    return (
        A_avg_homoatomic_dist_by_shortest_dist,
        B_avg_homoatomic_dist_by_shortest_dist,
        C_avg_homoatomic_dist_by_shortest_dist,
        D_avg_homoatomic_dist_by_shortest_dist,
    )


def get_A_and_B_and_C_and_D_count_in_best_label_per_element(connections, A, B, C, D):
    """Compute the occurrences of element types A, B, C, D at the
    shortest distances within their respective best labeled site for
    each element.

    The best site label is determined with the label with the shortest
    distance pair.
    """
    A_count_at_A_shortest_dist = 0
    A_count_at_B_shortest_dist = 0
    A_count_at_C_shortest_dist = 0
    A_count_at_D_shortest_dist = 0
    B_count_at_A_shortest_dist = 0
    B_count_at_B_shortest_dist = 0
    B_count_at_C_shortest_dist = 0
    B_count_at_D_shortest_dist = 0
    C_count_at_A_shortest_dist = 0
    C_count_at_B_shortest_dist = 0
    C_count_at_C_shortest_dist = 0
    C_count_at_D_shortest_dist = 0
    D_count_at_A_shortest_dist = 0
    D_count_at_B_shortest_dist = 0
    D_count_at_C_shortest_dist = 0
    D_count_at_D_shortest_dist = 0
    dist_count_per_label = count_first_second_min_dist(connections)
    best_label_dict = extract_best_labels(dist_count_per_label)
    best_A_site_label = best_label_dict[A]["best_label"]
    best_A_shortest_dist = best_label_dict[A]["best_label_details"]["shortest_dist"]
    best_B_site_label = best_label_dict[B]["best_label"]
    best_B_shortest_dist = best_label_dict[B]["best_label_details"]["shortest_dist"]
    best_C_site_label = best_label_dict[C]["best_label"]
    best_C_shortest_dist = best_label_dict[C]["best_label_details"]["shortest_dist"]
    best_D_site_label = best_label_dict[D]["best_label"]
    best_D_shortest_dist = best_label_dict[D]["best_label_details"]["shortest_dist"]
    for connection in connections[best_A_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_A_shortest_dist:
            if other_label.startswith(A):
                A_count_at_A_shortest_dist += 1
            if other_label.startswith(B):
                B_count_at_A_shortest_dist += 1
            if other_label.startswith(C):
                C_count_at_A_shortest_dist += 1
            if other_label.startswith(D):
                D_count_at_A_shortest_dist += 1
    for connection in connections[best_B_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_B_shortest_dist:
            if other_label.startswith(A):
                A_count_at_B_shortest_dist += 1
            if other_label.startswith(B):
                B_count_at_B_shortest_dist += 1
            if other_label.startswith(C):
                C_count_at_B_shortest_dist += 1
            if other_label.startswith(D):
                D_count_at_B_shortest_dist += 1
    for connection in connections[best_C_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_C_shortest_dist:
            if other_label.startswith(A):
                A_count_at_C_shortest_dist += 1
            if other_label.startswith(B):
                B_count_at_C_shortest_dist += 1
            if other_label.startswith(C):
                C_count_at_C_shortest_dist += 1
            if other_label.startswith(D):
                D_count_at_C_shortest_dist += 1
    for connection in connections[best_D_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_D_shortest_dist:
            if other_label.startswith(A):
                A_count_at_D_shortest_dist += 1
            if other_label.startswith(B):
                B_count_at_D_shortest_dist += 1
            if other_label.startswith(C):
                C_count_at_D_shortest_dist += 1
            if other_label.startswith(D):
                D_count_at_D_shortest_dist += 1
    return (
        A_count_at_A_shortest_dist,
        A_count_at_B_shortest_dist,
        A_count_at_C_shortest_dist,
        A_count_at_D_shortest_dist,
        B_count_at_A_shortest_dist,
        B_count_at_B_shortest_dist,
        B_count_at_C_shortest_dist,
        B_count_at_D_shortest_dist,
        C_count_at_A_shortest_dist,
        C_count_at_B_shortest_dist,
        C_count_at_C_shortest_dist,
        C_count_at_D_shortest_dist,
        D_count_at_A_shortest_dist,
        D_count_at_B_shortest_dist,
        D_count_at_C_shortest_dist,
        D_count_at_D_shortest_dist,
    )


def get_avg_A_and_B_and_C_and_D_count_in_per_element(connections, A, B, C, D):
    """Compute the occurrences of element types A, B, C, D at the
    shortest distances within their respective site for all site
    labels."""
    A_total_count_at_A_shortest_dist = 0
    A_total_count_at_B_shortest_dist = 0
    A_total_count_at_C_shortest_dist = 0
    A_total_count_at_D_shortest_dist = 0
    B_total_count_at_A_shortest_dist = 0
    B_total_count_at_B_shortest_dist = 0
    B_total_count_at_C_shortest_dist = 0
    B_total_count_at_D_shortest_dist = 0
    C_total_count_at_A_shortest_dist = 0
    C_total_count_at_B_shortest_dist = 0
    C_total_count_at_C_shortest_dist = 0
    C_total_count_at_D_shortest_dist = 0
    D_total_count_at_A_shortest_dist = 0
    D_total_count_at_B_shortest_dist = 0
    D_total_count_at_C_shortest_dist = 0
    D_total_count_at_D_shortest_dist = 0
    dist_count_per_label = count_first_second_min_dist(connections)
    best_labels = extract_best_labels(dist_count_per_label)
    A_site_label_count = best_labels[A]["label_count"]
    B_site_label_count = best_labels[B]["label_count"]
    C_site_label_count = best_labels[C]["label_count"]
    D_site_label_count = best_labels[D]["label_count"]

    for site_label, connection_data in connections.items():
        parsed_element = string_parser.get_atom_type_from_label(site_label)
        shortest_dist_per_site_label = dist_count_per_label[site_label]["shortest_dist"]

        for connection in connection_data:
            other_label, dist, _, _ = connection
            if parsed_element == A and dist == shortest_dist_per_site_label:
                if other_label.startswith(A):
                    A_total_count_at_A_shortest_dist += 1
                if other_label.startswith(B):
                    B_total_count_at_A_shortest_dist += 1
                if other_label.startswith(C):
                    C_total_count_at_A_shortest_dist += 1
                if other_label.startswith(D):
                    D_total_count_at_A_shortest_dist += 1
            if parsed_element == B and dist == shortest_dist_per_site_label:
                if other_label.startswith(A):
                    A_total_count_at_B_shortest_dist += 1
                if other_label.startswith(B):
                    B_total_count_at_B_shortest_dist += 1
                if other_label.startswith(C):
                    C_total_count_at_B_shortest_dist += 1
                if other_label.startswith(D):
                    D_total_count_at_B_shortest_dist += 1
            if parsed_element == C and dist == shortest_dist_per_site_label:
                if other_label.startswith(A):
                    A_total_count_at_C_shortest_dist += 1
                if other_label.startswith(B):
                    B_total_count_at_C_shortest_dist += 1
                if other_label.startswith(C):
                    C_total_count_at_C_shortest_dist += 1
                if other_label.startswith(D):
                    D_total_count_at_C_shortest_dist += 1
            if parsed_element == D and dist == shortest_dist_per_site_label:
                if other_label.startswith(A):
                    A_total_count_at_D_shortest_dist += 1
                if other_label.startswith(B):
                    B_total_count_at_D_shortest_dist += 1
                if other_label.startswith(C):
                    C_total_count_at_D_shortest_dist += 1
                if other_label.startswith(D):
                    D_total_count_at_D_shortest_dist += 1

    A_avg_count_at_A = A_total_count_at_A_shortest_dist / A_site_label_count
    A_avg_count_at_B = A_total_count_at_B_shortest_dist / B_site_label_count
    A_avg_count_at_C = A_total_count_at_C_shortest_dist / C_site_label_count
    A_avg_count_at_D = A_total_count_at_D_shortest_dist / D_site_label_count
    B_avg_count_at_A = B_total_count_at_A_shortest_dist / A_site_label_count
    B_avg_count_at_B = B_total_count_at_B_shortest_dist / B_site_label_count
    B_avg_count_at_C = B_total_count_at_C_shortest_dist / C_site_label_count
    B_avg_count_at_D = B_total_count_at_D_shortest_dist / D_site_label_count
    C_avg_count_at_A = C_total_count_at_A_shortest_dist / A_site_label_count
    C_avg_count_at_B = C_total_count_at_B_shortest_dist / B_site_label_count
    C_avg_count_at_C = C_total_count_at_C_shortest_dist / C_site_label_count
    C_avg_count_at_D = C_total_count_at_D_shortest_dist / D_site_label_count
    D_avg_count_at_A = D_total_count_at_A_shortest_dist / A_site_label_count
    D_avg_count_at_B = D_total_count_at_B_shortest_dist / B_site_label_count
    D_avg_count_at_C = D_total_count_at_C_shortest_dist / C_site_label_count
    D_avg_count_at_D = D_total_count_at_D_shortest_dist / D_site_label_count

    return (
        A_avg_count_at_A,
        A_avg_count_at_B,
        A_avg_count_at_C,
        A_avg_count_at_D,
        B_avg_count_at_A,
        B_avg_count_at_B,
        B_avg_count_at_C,
        B_avg_count_at_D,
        C_avg_count_at_A,
        C_avg_count_at_B,
        C_avg_count_at_C,
        C_avg_count_at_D,
        D_avg_count_at_A,
        D_avg_count_at_B,
        D_avg_count_at_C,
        D_avg_count_at_D,
    )



from cifkit import Cif

from SAF.features.wyc import helper


def compute_features(cif: Cif, elements):
    A, B, C, D = elements
    A_env, B_env, C_env, D_env = helper.get_quaternary_site_info(cif._loop_values, A, B, C, D)
    A_sites_total = A_env["sites"]
    B_sites_total = B_env["sites"]
    C_sites_total = C_env["sites"]
    D_sites_total = D_env["sites"]
    A_multiplicity_total = A_env["multiplicity"]
    B_multiplicity_total = B_env["multiplicity"]
    C_multiplicity_total = C_env["multiplicity"]
    D_multiplicity_total = D_env["multiplicity"]
    A_lowest_wyckoff_multiplicity = A_env["lowest_wyckoff_multiplicity"]
    B_lowest_wyckoff_multiplicity = B_env["lowest_wyckoff_multiplicity"]
    C_lowest_wyckoff_multiplicity = C_env["lowest_wyckoff_multiplicity"]
    D_lowest_wyckoff_multiplicity = D_env["lowest_wyckoff_multiplicity"]
    # Determine the lowest Wyckoff multiplicity
    min_wyckoff_multiplicity = min(
        A_lowest_wyckoff_multiplicity,
        B_lowest_wyckoff_multiplicity,
        C_lowest_wyckoff_multiplicity,
        D_lowest_wyckoff_multiplicity,
    )
    # Identify elements with the lowest Wyckoff multiplicity
    lowest_wyckoff_elements = []
    if A_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(A)
    if B_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(B)
    if C_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(C)
    if D_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(D)
    identical_lowest_wyckoff_multiplicity_count = len(lowest_wyckoff_elements)
    data = {
        "WYK_A_lowest_wyckoff": A_lowest_wyckoff_multiplicity,
        "WYK_B_lowest_wyckoff": B_lowest_wyckoff_multiplicity,
        "WYK_C_lowest_wyckoff": C_lowest_wyckoff_multiplicity,
        "WYK_D_lowest_wyckoff": D_lowest_wyckoff_multiplicity,
        "WYK_identical_lowest_wyckoff_count": identical_lowest_wyckoff_multiplicity_count,
        "WYK_A_sites_total": A_sites_total,
        "WYK_B_sites_total": B_sites_total,
        "WYK_C_sites_total": C_sites_total,
        "WYK_D_sites_total": D_sites_total,
        "WYK_A_multiplicity_total": A_multiplicity_total,
        "WYK_B_multiplicity_total": B_multiplicity_total,
        "WYK_C_multiplicity_total": C_multiplicity_total,
        "WYK_D_multiplicity_total": D_multiplicity_total,
    }
    uni_data = {
        "UNI_WYK_lowest_wyckoff": min_wyckoff_multiplicity,
    }
    return data, uni_data



#!/usr/bin/env python
##############################################################################
#
# (c) 2025 Sangjoon Lee.
# All rights reserved.
#
# File coded by: Sangjoon Lee, Anton Oliynyk, Emil Jaffal, Danila Shiryaev.
#
# See GitHub contributions for a more detailed list of contributors.
# https://github.com/bobleesj/structure-analyzer-featurizer/graphs/contributors
#
# See LICENSE.rst for license information.
#
##############################################################################
"""Python package to generate geometric features of interatomic
distances, atomic environment information, and coordination numbers from
a folder containing .cif files."""

# package version
from SAF.version import __version__  # noqa

# silence the pyflakes syntax checker
assert __version__ or True

# End of file



from collections import defaultdict

import numpy as np
from cifkit.utils import string_parser


def count_first_second_min_dist(
    all_labels_connections,
):
    """Determine the shortest and second shortest distances from each
    site label and their respective counts.

    Examples
    --------
    >>> connections = {
    ...     "Rh1": [("Rh1", 2.0, "Rh1", "Rh1"), ("Rh1", 3.0, "Rh1", "Rh2")],
    ...     "Rh2": [("Rh2", 1.5, "Rh2", "Rh1"), ("Rh2", 2.5, "Rh2", "Rh2")],
    ... }
    >>> result = count_first_second_min_dist(connections)
    >>> print(result)
    {"Rh1": {"shortest_dist": 2.0, "second_shortest_dist": 3.0, "counts": {2.0: 1, 3.0: 1}},
     "Rh2": {"shortest_dist": 1.5, "second_shortest_dist": 2.5, "counts": {1.5: 1, 2.5: 1}}}
    """
    # Dictionary to hold the results
    first_second_dist_info = defaultdict(
        lambda: {
            "shortest_dist": None,
            "second_shortest_dist": None,
            "counts": {},
        }
    )
    # Collect all distances for each label
    for (
        label,
        connections,
    ) in all_labels_connections.items():
        distances = []
        for _, dist, _, _ in connections:
            distances.append(dist)
        if distances:
            sorted_distances = sorted(distances)
            unique_distances, counts = np.unique(sorted_distances, return_counts=True)
            # Populate the shortest_dist and second shortest_dist distances
            if len(unique_distances) > 0:
                first_second_dist_info[label]["shortest_dist"] = float(unique_distances[0])
                first_second_dist_info[label]["counts"][float(unique_distances[0])] = int(counts[0])
            if len(unique_distances) > 1:
                first_second_dist_info[label]["second_shortest_dist"] = float(unique_distances[1])
                first_second_dist_info[label]["counts"][float(unique_distances[1])] = int(counts[1])
    return dict(first_second_dist_info)


def extract_best_labels(data):
    """Extract the best label for each element based on the shortest
    distance, and if there's a tie, use the second shortest distance as
    a tiebreaker.

    Attributes
    ----------
    data : first_second_dist_per_site_data

    Examples
    --------
    >>> data = {
    ...     "In1": {"shortest_dist": 2.697, "second_shortest_dist": 2.852, "counts": {2.697: 2, 2.852: 2}},
    ...     "U1": {"shortest_dist": 2.983, "second_shortest_dist": 2.984, "counts": {2.983: 3, 2.984: 1}},
    ...     "Rh1": {"shortest_dist": 2.852, "second_shortest_dist": 2.853, "counts": {2.852: 2, 2.853: 1}},
    ...     "Rh2": {"shortest_dist": 2.697, "second_shortest_dist": 3.046, "counts": {2.697: 6, 3.046: 3}},
    ...     }
    >>> result = extract_best_labels(data)
    >>> print(result)
    {
        "In": {
            "label_count": 1,
            "shortest_dist_total_count": 2,
            "second_shortest_dist_count": 2,
            "avg_shortest_dist": 2.0,
            "avg_second_shortest_dist_count": 2.0,
            "best_label": "In1",
            "best_label_details": {"shortest_dist": 2.697, "second_shortest_dist": 2.852, "counts": {2.697: 2, 2.852: 2}},
        },
        "U": {
            "label_count": 1,
            "shortest_dist_total_count": 3,
            "second_shortest_dist_count": 1,
            "avg_shortest_dist": 3.0,
            "avg_second_shortest_dist_count": 1.0,
            "best_label": "U1",
            "best_label_details": {"shortest_dist": 2.983, "second_shortest_dist": 2.984, "counts": {2.983: 3, 2.984: 1}},
        },
        "Rh": {
            "label_count": 2,
            "shortest_dist_total_count": 8,
            "second_shortest_dist_count": 4,
            "avg_shortest_dist": 4.0,
            "avg_second_shortest_dist_count": 2.0,
            "best_label": "Rh2",
            "best_label_details": {"shortest_dist": 2.697, "second_shortest_dist": 3.046, "counts": {2.697: 6, 3.046: 3}},
        },
    }
    """
    best_labels = {}
    label_counts = {}
    total_distances = {}
    # Initialize storage for distance statistics
    for label in data:
        element = string_parser.get_atom_type_from_label(label)  # Assuming label ends with a number like 'Rh1', 'Th1', etc.
        if element not in total_distances:
            total_distances[element] = {
                "shortest_counts": [],
                "second_shortest_counts": [],
            }
    # Collect data for each label
    for label, details in data.items():
        element = string_parser.get_atom_type_from_label(label)
        # Update label counts
        label_counts[element] = label_counts.get(element, 0) + 1
        # Update total distance counts
        total_distances[element]["shortest_counts"].append(details["counts"][details["shortest_dist"]])
        if "second_shortest_dist" in details:
            total_distances[element]["second_shortest_counts"].append(details["counts"].get(details["second_shortest_dist"], 0))
        # Determine best label
        if element not in best_labels:
            best_labels[element] = (label, details)
        else:
            current_best = best_labels[element][1]
            if (details["shortest_dist"] < current_best["shortest_dist"]) or (
                details["shortest_dist"] == current_best["shortest_dist"]
                and details["second_shortest_dist"] < current_best["second_shortest_dist"]
            ):
                best_labels[element] = (label, details)
    # Prepare final output
    output = {}
    for element, (label, details) in best_labels.items():
        total_shortest = sum(total_distances[element]["shortest_counts"])
        total_second_shortest = sum(total_distances[element]["second_shortest_counts"])
        avg_shortest = total_shortest / label_counts[element]
        avg_second_shortest = total_second_shortest / label_counts[element] if total_distances[element]["second_shortest_counts"] else 0
        output[element] = {
            "label_count": label_counts[element],
            "shortest_dist_total_count": total_shortest,
            "second_shortest_dist_count": total_second_shortest,
            "avg_shortest_dist": avg_shortest,
            "avg_second_shortest_dist_count": avg_second_shortest,
            "best_label": label,
            "best_label_details": details,
        }
    return output


def extract_shortest_dist_with_tol(best_label_dist_info, connections, tol=0.05):
    """Compute the count of distances within a specified tolerance of
    the shortest distance of each site label."""
    best_site_result = {}

    # Find the shortest distance form the best site label
    for element, info in best_label_dist_info.items():
        best_site_shortest_dist_count_within_tol = 0
        best_site_min_dist = best_label_dist_info[element]["best_label_details"]["shortest_dist"]
        best_site_label = best_label_dist_info[element]["best_label"]
        best_site_min_dist_with_tol = best_site_min_dist * (1 + tol)
        for connection in connections[best_site_label]:
            _, dist, _, _ = connection
            if dist <= best_site_min_dist_with_tol:
                best_site_shortest_dist_count_within_tol += 1
        best_site_result[element] = {
            "shortest_dist_count_within_tol": best_site_shortest_dist_count_within_tol,
        }
    return best_site_result


def extract_avg_shortest_dist_with_tol(connection_data, tol=0.05):
    """Compute the count of distances within a specified tolerance of
    the shortest distance of each site label."""
    site_result = defaultdict(lambda: {"shortest_dist_count_within_tol": 0})
    for site_label, connections in connection_data.items():
        site_result[site_label] = {}
        count_within_tol = 0
        min_dist_per_site_label = connections[0][1]
        site_min_dist_with_tol = min_dist_per_site_label * (1 + tol)
        for connection in connections:
            dist = connection[1]
            if dist <= site_min_dist_with_tol:
                count_within_tol += 1
        site_result[site_label] = {"shortest_dist_count_within_tol": count_within_tol}
    # Step 2: Aggregate counts by element and compute averages
    avg_result = defaultdict(lambda: {"total_shortest_dist_within_tol_count": 0})
    element_to_site_labels = _get_site_labels_per_element(connection_data)
    for site_label, result in site_result.items():
        element = string_parser.get_atom_type_from_label(site_label)
        if site_label in element_to_site_labels[element]:
            avg_result[element]["total_shortest_dist_within_tol_count"] += result["shortest_dist_count_within_tol"]
    for element, result in avg_result.items():
        num_sites = len(element_to_site_labels[element])
        if num_sites > 0:
            result["avg_shortest_dist_within_tol_count"] = result["total_shortest_dist_within_tol_count"] / num_sites
        else:
            result["avg_shortest_dist_within_tol_count"] = 0.0
    return avg_result


def get_avg_second_by_first_shortest_dist_ratio(site_dist_info, connection_data):
    """Compute the ratio of the second shortest distance to the shortest
    distance for each site label."""
    ratio_total_result = defaultdict(lambda: {"total_second_by_first_shortest_dist": 0.0})
    for site_label, data in site_dist_info.items():
        element = string_parser.get_atom_type_from_label(site_label)
        ratio = data["second_shortest_dist"] / data["shortest_dist"]
        ratio_total_result[element]["total_second_by_first_shortest_dist"] += ratio
    element_to_site_labels = _get_site_labels_per_element(connection_data)
    ratio_avg_result = {}
    for element, result in ratio_total_result.items():
        num_sites = len(element_to_site_labels[element])
        if num_sites > 0:
            avg_ratio = result["total_second_by_first_shortest_dist"] / num_sites
        else:
            avg_ratio = 0.0
        ratio_avg_result[element] = {"avg_second_by_first_shortest_dist": avg_ratio}
    return ratio_avg_result


def _get_site_labels_per_element(connection_data):
    element_to_site_labels = {}
    for site_label, _ in connection_data.items():
        element_label = string_parser.get_atom_type_from_label(site_label)
        if element_label not in element_to_site_labels:
            element_to_site_labels[element_label] = []
        element_to_site_labels[element_label].append(site_label)
    return element_to_site_labels



from cifkit import Cif

from SAF.features.wyc import helper


def compute_features(cif: Cif, elements):
    R, M, X = elements
    R_env, M_env, X_env = helper.get_ternary_site_info(cif._loop_values, R, M, X)
    R_sites_total = R_env["sites"]
    M_sites_total = M_env["sites"]
    X_sites_total = X_env["sites"]
    R_multiplicity_total = R_env["multiplicity"]
    M_multiplicity_total = M_env["multiplicity"]
    X_multiplicity_total = X_env["multiplicity"]
    M_lowest_wyckoff_multiplicity = M_env["lowest_wyckoff_multiplicity"]
    R_lowest_wyckoff_multiplicity = R_env["lowest_wyckoff_multiplicity"]
    X_lowest_wyckoff_multiplicity = X_env["lowest_wyckoff_multiplicity"]
    # Create a list to store the elements with the lowest Wyckoff label
    lowest_wyckoff_elements = []
    # Determine the lowest Wyckoff label between A and B
    min_wyckoff_multiplicity = min(
        R_lowest_wyckoff_multiplicity,
        M_lowest_wyckoff_multiplicity,
        X_lowest_wyckoff_multiplicity,
    )
    # If R, M. X the lowest Wyckoff label, add them to the list
    if R_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(R)
    if M_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(M)
    if X_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(X)
    identical_lowest_wyckoff_multiplicity_count = len(lowest_wyckoff_elements)
    data = {
        "WYK_R_lowest_wyckoff": R_lowest_wyckoff_multiplicity,
        "WYK_M_lowest_wyckoff": M_lowest_wyckoff_multiplicity,
        "WYK_X_lowest_wyckoff": X_lowest_wyckoff_multiplicity,
        "WYK_identical_lowest_wyckoff_count": identical_lowest_wyckoff_multiplicity_count,
        "WYK_R_sites_total": R_sites_total,
        "WYK_M_sites_total": M_sites_total,
        "WYK_X_sites_total": X_sites_total,
        "WYK_R_multiplicity_total": R_multiplicity_total,
        "WYK_M_multiplicity_total": M_multiplicity_total,
        "WYK_X_multiplicity_total": X_multiplicity_total,
    }
    uni_data = {
        "UNI_WYK_lowest_wyckoff": min_wyckoff_multiplicity,
    }
    return data, uni_data



import pytest

from SAF.features.coordination.ternary import compute_features


def test_features(URhIn_cif, URhIn_elements_tuple):
    result = compute_features(URhIn_cif, URhIn_elements_tuple)
    expected = {
        "CN_AVG_coordination_number": 11.5,
        "CN_AVG_R_atom_count": 4.5,
        "CN_AVG_M_atom_count": 2.25,
        "CN_AVG_X_atom_count": 4.75,
        "CN_AVG_polyhedron_volume": 67.79050000000001,
        "CN_AVG_central_atom_to_center_of_mass_dist": 0.07925000000000001,
        "CN_AVG_number_of_edges": 28.5,
        "CN_AVG_number_of_faces": 19.0,
        "CN_AVG_shortest_distance_to_face": 2.158375,
        "CN_AVG_shortest_distance_to_edge": 2.238625,
        "CN_AVG_volume_of_inscribed_sphere": 44.27,
        "CN_AVG_packing_efficiency": 0.64575,
        "CN_MIN_coordination_number": 10.75,
        "CN_MIN_R_atom_count": 3.75,
        "CN_MIN_M_atom_count": 2.25,
        "CN_MIN_X_atom_count": 4.75,
        "CN_MIN_polyhedron_volume": 60.91375,
        "CN_MIN_central_atom_to_center_of_mass_dist": 0.04225,
        "CN_MIN_number_of_edges": 26.25,
        "CN_MIN_number_of_faces": 17.5,
        "CN_MIN_shortest_distance_to_face": 2.0749999999999997,
        "CN_MIN_shortest_distance_to_edge": 2.14975,
        "CN_MIN_volume_of_inscribed_sphere": 38.59575,
        "CN_MIN_packing_efficiency": 0.62875,
        "CN_MAX_coordination_number": 12.25,
        "CN_MAX_R_atom_count": 5.25,
        "CN_MAX_M_atom_count": 2.25,
        "CN_MAX_X_atom_count": 4.75,
        "CN_MAX_polyhedron_volume": 74.66725,
        "CN_MAX_central_atom_to_center_of_mass_dist": 0.11625,
        "CN_MAX_number_of_edges": 30.75,
        "CN_MAX_number_of_faces": 20.5,
        "CN_MAX_shortest_distance_to_face": 2.2417499999999997,
        "CN_MAX_shortest_distance_to_edge": 2.3274999999999997,
        "CN_MAX_volume_of_inscribed_sphere": 49.94425,
        "CN_MAX_packing_efficiency": 0.66275,
    }
    for key in result:
        assert result[key] == pytest.approx(expected[key], abs=0.01)



import numpy as np
import pytest

from SAF.features.interatomic.quaternary import compute_features


def test_compute_features(Tb4RhInGe4_cif, TbRhInGe_elements_tuple):
    actual, _ = compute_features(Tb4RhInGe4_cif, True, TbRhInGe_elements_tuple)
    expected = {
        "Entry": "Tb4RhInGe4",
        "Formula": "Tb4RhInGe4",
        "Structure": "Tb4RhInGe4",
        "A": "Tb",
        "B": "Rh",
        "C": "In",
        "D": "Ge",
        "INT_AA_dist": 3.538,
        "INT_BB_dist": 3.148,
        "INT_CC_dist": 4.264,
        "INT_DD_dist": 2.596,
        "INT_AB_dist": 3.076,
        "INT_AC_dist": 3.367,
        "INT_AD_dist": 2.917,
        "INT_BC_dist": 4.711,
        "INT_BD_dist": 2.47,
        "INT_CD_dist": 2.818,
        "INT_Asize": 1.764,
        "INT_Bsize": 1.345,
        "INT_Csize": 1.624,
        "INT_Dsize": 1.225,
        "INT_Asize_by_Bsize": 1.3115241635687733,
        "INT_Bsize_by_Csize": 0.8282019704433496,
        "INT_Csize_by_Dsize": 1.3257142857142856,
        "INT_Asize_by_Csize": 1.086206896551724,
        "INT_Asize_by_Dsize": 1.44,
        "INT_Bsize_by_Dsize": 1.0979591836734692,
        "INT_AA_dist_by_2_by_Asize": 1.0028344671201814,
        "INT_BB_dist_by_2_by_Bsize": 1.1702602230483272,
        "INT_CC_dist_by_2_by_Csize": 1.312807881773399,
        "INT_DD_dist_by_2_by_Dsize": 1.059591836734694,
        "INT_AB_dist_by_ABsizes": 0.9893856545513027,
        "INT_AC_dist_by_ACsizes": 0.9938016528925621,
        "INT_AD_dist_by_ADsizes": 0.9759116761458682,
        "INT_BC_dist_by_BCsizes": 1.5867295385651734,
        "INT_BD_dist_by_BDsizes": 0.961089494163424,
        "INT_CD_dist_by_CDsizes": 0.9891189891189891,
        "INT_Asize_ref": np.float64(1.6862073857087099),
        "INT_Bsize_ref": np.float64(1.4201916565448087),
        "INT_Csize_ref": np.float64(1.6861073857088038),
        "INT_Dsize_ref": np.float64(1.1604492903475334),
        "INT_percent_diff_A_by_100": np.float64(-0.04410012148032321),
        "INT_percent_diff_B_by_100": np.float64(0.05590457735673513),
        "INT_percent_diff_C_by_100": np.float64(0.038243464106406215),
        "INT_percent_diff_D_by_100": np.float64(-0.05269445685915647),
        "INT_AA_minus_ref_diff": np.float64(0.04680193006856418),
        "INT_BB_minus_ref_diff": np.float64(0.0977181343425612),
        "INT_CC_minus_ref_diff": np.float64(0.20914287724727781),
        "INT_DD_minus_ref_diff": np.float64(0.10597127091869539),
        "INT_AB_minus_ref_diff": np.float64(-0.009882653528452056),
        "INT_AC_minus_ref_diff": np.float64(-0.0015784886894902542),
        "INT_AD_minus_ref_diff": np.float64(0.02411495507156549),
        "INT_BC_minus_ref_diff": np.float64(0.3406285200056013),
        "INT_BD_minus_ref_diff": np.float64(-0.04479390562442993),
        "INT_CD_minus_ref_diff": np.float64(-0.010133667869530575),
        "INT_R_factor": np.float64(0.00930941081455393),
        "INT_UNI_shortest_homoatomic_dist": 2.596,
        "INT_UNI_shortest_heteroatomic_dist": 2.47,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 1.059591836734694,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 0.961089494163424,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": np.float64(1.1185322881375306),
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": np.float64(0.9571265630634986),
        "INT_UNI_highest_refined_percent_diff_abs": np.float64(0.05590457735673513),
        "INT_UNI_lowest_refined_percent_diff_abs": np.float64(0.038243464106406215),
        "INT_UNI_refined_packing_efficiency": np.float64(0.64862),
    }
    for key, expected_value in expected.items():
        # For float
        if isinstance(expected_value, float):
            assert actual[key] == pytest.approx(expected_value, abs=0.001)
        # For str, integer, etc.
        else:
            assert actual[key] == expected_value



from bobleesj.utils.sorters.element_sorter import ElementSorter
from cifkit import Cif

from SAF.features.coordination import binary as CN_binary
from SAF.features.coordination import quaternary as CN_quaternary
from SAF.features.coordination import ternary as CN_ternary
from SAF.features.environment import binary as env_binary
from SAF.features.environment import quaternary as env_quaternary
from SAF.features.environment import ternary as env_ternary
from SAF.features.interatomic import binary as int_binary
from SAF.features.interatomic import quaternary as int_quaternary
from SAF.features.interatomic import ternary as int_ternary
from SAF.features.wyc import binary as wyc_binary
from SAF.features.wyc import quaternary as wyc_quaternary
from SAF.features.wyc import ternary as wyc_ternary


def _generate_features(
    file_path: str, supercell_size: int, use_size_constraint: bool, custom_labels, int_module, wyc_module, env_module, CN_module
):
    cif = Cif(file_path, supercell_size=supercell_size, compute_CN=True)
    elements = list(cif.unique_elements)
    element_sorter = ElementSorter(label_mapping=custom_labels)
    if custom_labels:
        elements_sorted = element_sorter.sort(elements, method="custom")
    else:
        elements_sorted = element_sorter.sort(elements, method="mendeleev")
    elements_tuple = tuple(elements_sorted)
    int_feat, int_uni = int_module.compute_features(cif, use_size_constraint, elements_tuple)
    wyc_feat, wyc_uni = wyc_module.compute_features(cif, elements_sorted)
    env_feat = env_module.compute_features(cif, elements_sorted)
    CN_feat = CN_module.compute_features(cif, elements_sorted)
    features = {
        **int_feat,
        **wyc_feat,
        **env_feat,
        **CN_feat,
    }
    uni_features = {
        **int_uni,
        **wyc_uni,
    }
    return features, uni_features


def compute_binary_features(file_path: str, supercell_size=3, use_size_constraint=True, custom_labels=None):
    return _generate_features(
        file_path,
        supercell_size,
        use_size_constraint,
        custom_labels=custom_labels,
        int_module=int_binary,
        wyc_module=wyc_binary,
        env_module=env_binary,
        CN_module=CN_binary,
    )


def compute_ternary_features(file_path: str, supercell_size=3, use_size_constraint=True, custom_labels=None):
    return _generate_features(
        file_path,
        supercell_size,
        use_size_constraint,
        custom_labels=custom_labels,
        int_module=int_ternary,
        wyc_module=wyc_ternary,
        env_module=env_ternary,
        CN_module=CN_ternary,
    )


def compute_quaternary_features(file_path: str, supercell_size=3, use_size_constraint=True, custom_labels=None):
    return _generate_features(
        file_path,
        supercell_size,
        use_size_constraint,
        custom_labels=custom_labels,
        int_module=int_quaternary,
        wyc_module=wyc_quaternary,
        env_module=env_quaternary,
        CN_module=CN_quaternary,
    )



from deepdiff import DeepDiff

from SAF.features.generator import compute_ternary_features


def test_generate_features():
    file_path = "tests/cif/ternary/URhIn.cif"
    actual_features, actual_uni_features = compute_ternary_features(file_path)
    expected_features = {
        "Entry": "URhIn",
        "Formula": "URhIn",
        "Structure": "ZrNiAl",
        "R": "U",
        "M": "Rh",
        "X": "In",
        "INT_RR_dist": 3.881,
        "INT_MM_dist": 3.881,
        "INT_XX_dist": 3.244,
        "INT_RM_dist": 2.983,
        "INT_MX_dist": 2.697,
        "INT_RX_dist": 3.21,
        "INT_Rsize": 1.377,
        "INT_Msize": 1.345,
        "INT_Xsize": 1.624,
        "INT_Rsize_by_Msize": 1.0237918215613384,
        "INT_Msize_by_Xsize": 0.8282019704433496,
        "INT_Rsize_by_Xsize": 0.8479064039408867,
        "INT_RR_dist_by_2_by_Rsize": 1.4092229484386347,
        "INT_MM_dist_by_2_by_Msize": 1.4427509293680296,
        "INT_XX_dist_by_2_by_Xsize": 0.9987684729064039,
        "INT_RM_dist_by_RMsizes": 1.0958853783982367,
        "INT_MX_dist_by_MXsizes": 0.9083866621758167,
        "INT_RX_dist_by_RXsizes": 1.0696434521826057,
        "INT_Rsize_ref": 1.4874430151209064,
        "INT_Msize_ref": 1.3560056919905379,
        "INT_Xsize_ref": 1.4864270730120939,
        "INT_percent_diff_R_by_100": 0.08020553022578532,
        "INT_percent_diff_M_by_100": 0.008182670624935252,
        "INT_percent_diff_X_by_100": -0.0847123934654595,
        "INT_RR_dist_minus_ref_diff": 0.2334743544854901,
        "INT_MM_dist_minus_ref_diff": 0.3012080948257985,
        "INT_XX_dist_minus_ref_diff": 0.08358380208872147,
        "INT_RM_dist_minus_ref_diff": 0.04678219674440349,
        "INT_MX_dist_minus_ref_diff": -0.05392390248521745,
        "INT_RX_dist_minus_ref_diff": 0.07356072020778807,
        "INT_R_factor": 0.013676072784002367,
        "INT_UNI_shortest_homoatomic_dist": 3.244,
        "INT_UNI_shortest_heteroatomic_dist": 2.697,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 0.9987684729064039,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 0.9083866621758167,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": 1.09120725089673,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": 0.948835108153386,
        "INT_UNI_highest_refined_percent_diff_abs": 0.0847123934654595,
        "INT_UNI_lowest_refined_percent_diff_abs": 0.008182670624935252,
        "INT_UNI_refined_packing_efficiency": 0.606644672780686,
        "WYK_R_lowest_wyckoff": 3,
        "WYK_M_lowest_wyckoff": 1,
        "WYK_X_lowest_wyckoff": 3,
        "WYK_identical_lowest_wyckoff_count": 1,
        "WYK_R_sites_total": 1,
        "WYK_M_sites_total": 2,
        "WYK_X_sites_total": 1,
        "WYK_R_multiplicity_total": 3,
        "WYK_M_multiplicity_total": 3,
        "WYK_X_multiplicity_total": 3,
        "ENV_R_shortest_dist_count": 3,
        "ENV_M_shortest_dist_count": 6,
        "ENV_X_shortest_dist_count": 2,
        "ENV_R_avg_shortest_dist_count": 3.0,
        "ENV_M_avg_shortest_dist_count": 4.0,
        "ENV_X_avg_shortest_dist_count": 2.0,
        "ENV_R_shortest_tol_dist_count": 5,
        "ENV_M_shortest_tol_dist_count": 6,
        "ENV_X_shortest_tol_dist_count": 2,
        "ENV_R_avg_shortest_dist_within_tol_count": 5.0,
        "ENV_M_avg_shortest_dist_within_tol_count": 7.5,
        "ENV_X_avg_shortest_dist_within_tol_count": 2.0,
        "ENV_R_second_by_first_shortest_dist": 1.000335232986926,
        "ENV_M_second_by_first_shortest_dist": 1.1294030404152762,
        "ENV_X_second_by_first_shortest_dist": 1.057471264367816,
        "ENV_R_avg_second_by_first_shortest_dist": 1.000335232986926,
        "ENV_M_avg_second_by_first_shortest_dist": 1.0648768357756606,
        "ENV_X_avg_second_by_first_shortest_dist": 1.057471264367816,
        "ENV_R_second_shortest_dist_count": 1,
        "ENV_M_second_shortest_dist_count": 3,
        "ENV_X_second_shortest_dist_count": 2,
        "ENV_R_avg_second_shortest_dist_count": 1.0,
        "ENV_M_avg_second_shortest_dist_count": 2.0,
        "ENV_X_avg_second_shortest_dist_count": 2.0,
        "ENV_R_homoatomic_dist_by_shortest_dist": 1.3010392222594702,
        "ENV_M_homoatomic_dist_by_shortest_dist": 1.4390063032999627,
        "ENV_X_homoatomic_dist_by_shortest_dist": 1.2028179458657768,
        "ENV_R_avg_homoatomic_dist_by_shortest_dist": 1.3010392222594702,
        "ENV_M_avg_homoatomic_dist_by_shortest_dist": 1.3999028711450725,
        "ENV_X_avg_homoatomic_dist_by_shortest_dist": 1.2028179458657768,
        "ENV_R_count_at_R_shortest_dist": 0,
        "ENV_M_count_at_R_shortest_dist": 3,
        "ENV_X_count_at_R_shortest_dist": 0,
        "ENV_R_avg_count_at_R_shortest_dist": 0.0,
        "ENV_M_avg_count_at_R_shortest_dist": 0.0,
        "ENV_X_avg_count_at_R_shortest_dist": 0.0,
        "ENV_R_count_at_M_shortest_dist": 0,
        "ENV_M_count_at_M_shortest_dist": 0,
        "ENV_X_count_at_M_shortest_dist": 6,
        "ENV_R_avg_count_at_M_shortest_dist": 1.5,
        "ENV_M_avg_count_at_M_shortest_dist": 0.0,
        "ENV_X_avg_count_at_M_shortest_dist": 1.0,
        "ENV_R_count_at_X_shortest_dist": 0,
        "ENV_M_count_at_X_shortest_dist": 2,
        "ENV_X_count_at_X_shortest_dist": 0,
        "ENV_R_avg_count_at_X_shortest_dist": 0.0,
        "ENV_M_avg_count_at_X_shortest_dist": 8.0,
        "ENV_X_avg_count_at_X_shortest_dist": 0.0,
        "CN_AVG_coordination_number": 11.5,
        "CN_AVG_R_atom_count": 4.5,
        "CN_AVG_M_atom_count": 2.25,
        "CN_AVG_X_atom_count": 4.75,
        "CN_AVG_polyhedron_volume": 67.79050000000001,
        "CN_AVG_central_atom_to_center_of_mass_dist": 0.07925000000000001,
        "CN_AVG_number_of_edges": 28.5,
        "CN_AVG_number_of_faces": 19.0,
        "CN_AVG_shortest_distance_to_face": 2.16375,
        "CN_AVG_shortest_distance_to_edge": 2.238625,
        "CN_AVG_volume_of_inscribed_sphere": 44.545875,
        "CN_AVG_packing_efficiency": 0.650375,
        "CN_MIN_coordination_number": 10.75,
        "CN_MIN_R_atom_count": 3.75,
        "CN_MIN_M_atom_count": 2.25,
        "CN_MIN_X_atom_count": 4.75,
        "CN_MIN_polyhedron_volume": 60.91375,
        "CN_MIN_central_atom_to_center_of_mass_dist": 0.04225,
        "CN_MIN_number_of_edges": 26.25,
        "CN_MIN_number_of_faces": 17.5,
        "CN_MIN_shortest_distance_to_face": 2.08575,
        "CN_MIN_shortest_distance_to_edge": 2.14975,
        "CN_MIN_volume_of_inscribed_sphere": 39.1475,
        "CN_MIN_packing_efficiency": 0.638,
        "CN_MAX_coordination_number": 12.25,
        "CN_MAX_R_atom_count": 5.25,
        "CN_MAX_M_atom_count": 2.25,
        "CN_MAX_X_atom_count": 4.75,
        "CN_MAX_polyhedron_volume": 74.66725,
        "CN_MAX_central_atom_to_center_of_mass_dist": 0.11625,
        "CN_MAX_number_of_edges": 30.75,
        "CN_MAX_number_of_faces": 20.5,
        "CN_MAX_shortest_distance_to_face": 2.2417499999999997,
        "CN_MAX_shortest_distance_to_edge": 2.3274999999999997,
        "CN_MAX_volume_of_inscribed_sphere": 49.94425,
        "CN_MAX_packing_efficiency": 0.66275,
    }
    diff = DeepDiff(expected_features, actual_features, significant_digits=2)
    assert diff == {}
    expected_uni_features = {
        "Entry": "URhIn",
        "Formula": "URhIn",
        "INT_UNI_shortest_homoatomic_dist": 3.244,
        "INT_UNI_shortest_heteroatomic_dist": 2.697,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 0.9987684729064039,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 0.9083866621758167,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": 1.09120725089673,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": 0.948835108153386,
        "INT_UNI_highest_refined_percent_diff_abs": 0.0847123934654595,
        "INT_UNI_lowest_refined_percent_diff_abs": 0.008182670624935252,
        "INT_UNI_refined_packing_efficiency": 0.606644672780686,
        "UNI_WYK_lowest_wyckoff": 1,
    }
    diff = DeepDiff(expected_uni_features, actual_uni_features, significant_digits=2)
    assert diff == {}



from deepdiff import DeepDiff

from SAF.features.environment.quaternary import compute_features


def test_compute_binary_env_features_ThSb(Tb4RhInGe4_cif, TbRhInGe_elements_tuple):
    actual = compute_features(Tb4RhInGe4_cif, TbRhInGe_elements_tuple)
    expected = {
        "ENV_A_shortest_dist_count": 1,
        "ENV_B_shortest_dist_count": 2,
        "ENV_C_shortest_dist_count": 2,
        "ENV_D_shortest_dist_count": 2,
        "ENV_A_avg_shortest_dist_count": 1.75,
        "ENV_B_avg_shortest_dist_count": 2.0,
        "ENV_C_avg_shortest_dist_count": 2.0,
        "ENV_D_avg_shortest_dist_count": 1.25,
        "ENV_A_shortest_tol_dist_count": 6,
        "ENV_B_shortest_tol_dist_count": 4,
        "ENV_C_shortest_tol_dist_count": 4,
        "ENV_D_shortest_tol_dist_count": 3,
        "ENV_A_avg_shortest_dist_within_tol_count": 5.0,
        "ENV_B_avg_shortest_dist_within_tol_count": 4.0,
        "ENV_C_avg_shortest_dist_within_tol_count": 3.0,
        "ENV_D_avg_shortest_dist_within_tol_count": 1.75,
        "ENV_A_second_by_first_shortest_dist": 1.0205690778196779,
        "ENV_B_second_by_first_shortest_dist": 1.0291497975708501,
        "ENV_C_second_by_first_shortest_dist": 1.0184528034066713,
        "ENV_D_second_by_first_shortest_dist": 1.0291497975708501,
        "ENV_A_avg_second_by_first_shortest_dist": 1.0364821721769375,
        "ENV_B_avg_second_by_first_shortest_dist": 1.0291497975708501,
        "ENV_C_avg_second_by_first_shortest_dist": 1.0440696072782485,
        "ENV_D_avg_second_by_first_shortest_dist": 1.0680669343726787,
        "ENV_A_second_shortest_dist_count": 2,
        "EnV_B_second_shortest_dist_count": 1,
        "ENV_C_second_shortest_dist_count": 2,
        "ENV_D_second_shortest_dist_count": 1,
        "ENV_A_avg_second_shortest_dist_count": 2.0,
        "ENV_B_avg_second_shortest_dist_count": 1.0,
        "ENV_C_avg_second_shortest_dist_count": 2.0,
        "ENV_D_avg_second_shortest_dist_count": 1.25,
        "ENV_A_homoatomic_dist_by_shortest_dist": 1.2128899554336647,
        "ENV_B_homoatomic_dist_by_shortest_dist": 1.274493927125506,
        "ENV_C_homoatomic_dist_by_shortest_dist": 1.5131298793470547,
        "ENV_D_homoatomic_dist_by_shortest_dist": 1.0510121457489878,
        "ENV_A_avg_homoatomic_dist_by_shortest_dist": 1.215032553686275,
        "ENV_B_avg_homoatomic_dist_by_shortest_dist": 1.274493927125506,
        "ENV_C_avg_homoatomic_dist_by_shortest_dist": 1.4994220825306703,
        "ENV_D_avg_homoatomic_dist_by_shortest_dist": 1.0193165160610715,
        "ENV_A_count_at_A_shortest_dist": 0,
        "ENV_B_count_at_A_shortest_dist": 0,
        "ENV_C_count_at_A_shortest_dist": 0,
        "ENV_D_count_at_A_shortest_dist": 1,
        "ENV_A_avg_count_at_A_shortest_dist": 0.0,
        "ENV_B_avg_count_at_A_shortest_dist": 0.0,
        "ENV_C_avg_count_at_A_shortest_dist": 0.0,
        "ENV_D_avg_count_at_A_shortest_dist": 1.75,
        "ENV_A_count_at_B_shortest_dist": 0,
        "ENV_B_count_at_B_shortest_dist": 0,
        "ENV_C_count_at_B_shortest_dist": 0,
        "ENV_D_count_at_B_shortest_dist": 2,
        "ENV_A_avg_count_at_B_shortest_dist": 0.0,
        "ENV_B_avg_count_at_B_shortest_dist": 0.0,
        "ENV_C_avg_count_at_B_shortest_dist": 0.0,
        "ENV_D_avg_count_at_B_shortest_dist": 2.0,
        "ENV_A_count_at_C_shortest_dist": 0,
        "ENV_B_count_at_C_shortest_dist": 0,
        "ENV_C_count_at_C_shortest_dist": 0,
        "ENV_D_count_at_C_shortest_dist": 2,
        "ENV_A_avg_count_at_C_shortest_dist": 0.0,
        "ENV_B_avg_count_at_C_shortest_dist": 0.0,
        "ENV_C_avg_count_at_C_shortest_dist": 0.0,
        "ENV_D_avg_count_at_C_shortest_dist": 2.0,
        "ENV_A_count_at_D_shortest_dist": 0,
        "ENV_B_count_at_D_shortest_dist": 2,
        "ENV_C_count_at_D_shortest_dist": 0,
        "ENV_D_count_at_D_shortest_dist": 0,
        "ENV_A_avg_count_at_D_shortest_dist": 0.0,
        "ENV_B_avg_count_at_D_shortest_dist": 0.75,
        "ENV_C_avg_count_at_D_shortest_dist": 0.0,
        "ENV_D_avg_count_at_D_shortest_dist": 0.5,
    }
    diff = DeepDiff(actual, expected, significant_digits=3)
    assert diff == {}



import pytest

from SAF.features.interatomic.ternary import compute_features


def test_compute_ternary_interatomic_feature(URhIn_cif, URhIn_elements_tuple):
    actual, _ = compute_features(URhIn_cif, True, URhIn_elements_tuple)
    expected = {
        "Entry": "URhIn",
        "Formula": "URhIn",
        "Structure": "ZrNiAl",
        "R": "U",
        "M": "Rh",
        "X": "In",
        "INT_RR_dist": 3.881,
        "INT_MM_dist": 3.881,
        "INT_XX_dist": 3.244,
        "INT_RM_dist": 2.983,
        "INT_MX_dist": 2.697,
        "INT_RX_dist": 3.21,
        "INT_Rsize": 1.377,
        "INT_Msize": 1.345,
        "INT_Xsize": 1.624,
        "INT_Rsize_by_Msize": 1.0237918215613384,
        "INT_Msize_by_Xsize": 0.8282019704433496,
        "INT_Rsize_by_Xsize": 0.8479064039408867,
        "INT_RR_dist_by_2_by_Rsize": 1.4092229484386347,
        "INT_MM_dist_by_2_by_Msize": 1.4427509293680296,
        "INT_XX_dist_by_2_by_Xsize": 0.9987684729064039,
        "INT_RM_dist_by_RMsizes": 1.0958853783982367,
        "INT_MX_dist_by_MXsizes": 0.9083866621758167,
        "INT_RX_dist_by_RXsizes": 1.0696434521826057,
        "INT_Rsize_ref": 1.4874430151209064,
        "INT_Msize_ref": 1.3560056919905379,
        "INT_Xsize_ref": 1.4864270730120939,
        "INT_percent_diff_R_by_100": 0.08020553022578532,
        "INT_percent_diff_M_by_100": 0.008182670624935252,
        "INT_percent_diff_X_by_100": -0.0847123934654595,
        "INT_RR_dist_minus_ref_diff": 0.2334743544854901,
        "INT_MM_dist_minus_ref_diff": 0.3012080948257985,
        "INT_XX_dist_minus_ref_diff": 0.08358380208872147,
        "INT_RM_dist_minus_ref_diff": 0.04678219674440349,
        "INT_MX_dist_minus_ref_diff": -0.05392390248521745,
        "INT_RX_dist_minus_ref_diff": 0.07356072020778807,
        "INT_R_factor": 0.013676072784002367,
        "INT_UNI_shortest_homoatomic_dist": 3.244,
        "INT_UNI_shortest_heteroatomic_dist": 2.697,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 0.9987684729064039,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 0.9083866621758167,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": 1.09120725089673,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": 0.948835108153386,
        "INT_UNI_highest_refined_percent_diff_abs": 0.0847123934654595,
        "INT_UNI_lowest_refined_percent_diff_abs": 0.008182670624935252,
        "INT_UNI_refined_packing_efficiency": 0.60664,
    }
    for key, expected_value in expected.items():
        if isinstance(expected_value, float):
            assert actual[key] == pytest.approx(expected_value, abs=0.001)
        else:
            assert actual[key] == expected_value



from deepdiff import DeepDiff

from SAF.features.environment.ternary import compute_features


def test_compute_binary_env_features_ThSb(URhIn_cif, URhIn_elements_tuple):
    actual = compute_features(URhIn_cif, URhIn_elements_tuple)
    expected = {
        "ENV_R_shortest_dist_count": 3,
        "ENV_M_shortest_dist_count": 6,
        "ENV_X_shortest_dist_count": 2,
        "ENV_R_avg_shortest_dist_count": 3.0,
        "ENV_M_avg_shortest_dist_count": 4.0,
        "ENV_X_avg_shortest_dist_count": 2.0,
        "ENV_R_shortest_tol_dist_count": 5,
        "ENV_M_shortest_tol_dist_count": 6,
        "ENV_X_shortest_tol_dist_count": 2,
        "ENV_R_avg_shortest_dist_within_tol_count": 5.0,
        "ENV_M_avg_shortest_dist_within_tol_count": 7.5,
        "ENV_X_avg_shortest_dist_within_tol_count": 2.0,
        "ENV_R_second_by_first_shortest_dist": 1.000335232986926,
        "ENV_M_second_by_first_shortest_dist": 1.1294030404152762,
        "ENV_X_second_by_first_shortest_dist": 1.057471264367816,
        "ENV_R_avg_second_by_first_shortest_dist": 1.000335232986926,
        "ENV_M_avg_second_by_first_shortest_dist": 1.0648768357756606,
        "ENV_X_avg_second_by_first_shortest_dist": 1.057471264367816,
        "ENV_R_second_shortest_dist_count": 1,
        "ENV_M_second_shortest_dist_count": 3,
        "ENV_X_second_shortest_dist_count": 2,
        "ENV_R_avg_second_shortest_dist_count": 1.0,
        "ENV_M_avg_second_shortest_dist_count": 2.0,
        "ENV_X_avg_second_shortest_dist_count": 2.0,
        "ENV_R_homoatomic_dist_by_shortest_dist": 1.3010392222594702,
        "ENV_M_homoatomic_dist_by_shortest_dist": 1.4390063032999627,
        "ENV_X_homoatomic_dist_by_shortest_dist": 1.2028179458657768,
        "ENV_R_avg_homoatomic_dist_by_shortest_dist": 1.3010392222594702,
        "ENV_M_avg_homoatomic_dist_by_shortest_dist": 1.3999028711450725,
        "ENV_X_avg_homoatomic_dist_by_shortest_dist": 1.2028179458657768,
        "ENV_R_count_at_R_shortest_dist": 0,
        "ENV_M_count_at_R_shortest_dist": 3,
        "ENV_X_count_at_R_shortest_dist": 0,
        "ENV_R_avg_count_at_R_shortest_dist": 0.0,
        "ENV_M_avg_count_at_R_shortest_dist": 0.0,
        "ENV_X_avg_count_at_R_shortest_dist": 0.0,
        "ENV_R_count_at_M_shortest_dist": 0,
        "ENV_M_count_at_M_shortest_dist": 0,
        "ENV_X_count_at_M_shortest_dist": 6,
        "ENV_R_avg_count_at_M_shortest_dist": 1.5,
        "ENV_M_avg_count_at_M_shortest_dist": 0.0,
        "ENV_X_avg_count_at_M_shortest_dist": 1.0,
        "ENV_R_count_at_X_shortest_dist": 0,
        "ENV_M_count_at_X_shortest_dist": 2,
        "ENV_X_count_at_X_shortest_dist": 0,
        "ENV_R_avg_count_at_X_shortest_dist": 0.0,
        "ENV_M_avg_count_at_X_shortest_dist": 8.0,
        "ENV_X_avg_count_at_X_shortest_dist": 0.0,
    }
    diff = DeepDiff(actual, expected, significant_digits=3)
    assert diff == {}



from cifkit import Cif
from cifkit.data import radius_optimization as radius_opt

from SAF.features.interatomic import helper
from SAF.utils import bond, packing


def compute_features(cif: Cif, use_size_constraint: bool, elements):
    R, M, X = elements
    R_CIF_rad = cif.radius_values[R]["CIF_radius"]
    M_CIF_rad = cif.radius_values[M]["CIF_radius"]
    X_CIF_rad = cif.radius_values[X]["CIF_radius"]
    CIF_rad_refined, obj_value = radius_opt.get_refined_CIF_radius(
        [R, M, X], cif.shortest_bond_pair_distance, elements_ordered=False, use_size_constraint=use_size_constraint
    )
    R_CIF_rad_refined = float(CIF_rad_refined[R])
    M_CIF_rad_refined = float(CIF_rad_refined[M])
    X_CIF_rad_refined = float(CIF_rad_refined[X])
    # Do not use the alphabetical order of the elements for the refined radius
    min_bond_dists = bond.get_min_distances_by_labels(cif.shortest_bond_pair_distance, [R, M, X], labels=["R", "M", "X"])
    # 1) Min bond distances
    RR_dist = min_bond_dists["RR"]
    RM_dist = min_bond_dists["RM"]
    RX_dist = min_bond_dists["RX"]
    MM_dist = min_bond_dists["MM"]
    MX_dist = min_bond_dists["MX"]
    XX_dist = min_bond_dists["XX"]
    # 2) Radius from interatomic distances divided by CIF radius
    RM_dist_by_RM_sizes = RM_dist / (R_CIF_rad + M_CIF_rad)
    RX_dist_by_RX_sizes = RX_dist / (R_CIF_rad + X_CIF_rad)
    MX_dist_by_MX_sizes = MX_dist / (M_CIF_rad + X_CIF_rad)
    # 3) Radius from interatomic distances divided by 2 (3 of them for ternary)
    RR_dist_by_2 = RR_dist / 2
    MM_dist_by_2 = MM_dist / 2
    XX_dist_by_2 = XX_dist / 2
    # 4) Refined radius sums (6 of them for ternary)
    RR_ref_rad_sum = 2 * R_CIF_rad_refined
    MM_ref_rad_sum = 2 * M_CIF_rad_refined
    XX_ref_rad_sum = 2 * X_CIF_rad_refined
    RM_ref_rad_sum = R_CIF_rad_refined + M_CIF_rad_refined
    RX_ref_rad_sum = R_CIF_rad_refined + X_CIF_rad_refined
    MX_ref_rad_sum = M_CIF_rad_refined + X_CIF_rad_refined
    # 5) Percent differences between refined and CIF radius
    percent_diff_R = (R_CIF_rad_refined - R_CIF_rad) / R_CIF_rad
    percent_diff_M = (M_CIF_rad_refined - M_CIF_rad) / M_CIF_rad
    percent_diff_X = (X_CIF_rad_refined - X_CIF_rad) / X_CIF_rad
    # 6) Size ratios of CIF radius
    Rsize_by_Msize = R_CIF_rad / M_CIF_rad
    Msize_by_Xsize = M_CIF_rad / X_CIF_rad
    Rsize_by_Xsize = R_CIF_rad / X_CIF_rad
    # 7) Radius from interatomic distances divided by CIF radius
    RR_dist_by_2_by_Rsize = RR_dist_by_2 / R_CIF_rad
    MM_dist_by_2_by_Msize = MM_dist_by_2 / M_CIF_rad
    XX_dist_by_2_by_Xsize = XX_dist_by_2 / X_CIF_rad
    # 8) Difference between the iteratomic distance and the expected refined radius sum
    RR_minus_ref_diff = (RR_dist - RR_ref_rad_sum) / RR_dist
    MM_minus_ref_diff = (MM_dist - MM_ref_rad_sum) / MM_dist
    XX_minus_ref_diff = (XX_dist - XX_ref_rad_sum) / XX_dist
    RM_minus_ref_diff = (RM_dist - RM_ref_rad_sum) / RM_dist
    RX_minus_ref_diff = (RX_dist - RX_ref_rad_sum) / RX_dist
    MX_minus_ref_diff = (MX_dist - MX_ref_rad_sum) / MX_dist
    # 9) Get the shortest homo/hetroatomic distance
    homoatomic_dists = helper.get_bond_distances_by_labels(min_bond_dists, ["RR", "MM", "XX"])
    heteroatomic_dists = helper.get_bond_distances_by_labels(min_bond_dists, ["RM", "MX", "RX"])
    shortest_heteroatomic_dist = helper.get_min_from_dists(heteroatomic_dists)
    shortest_homoatomic_dist = helper.get_min_from_dists(homoatomic_dists)
    # Parse the CIF radius based on the shortest homo/heteroatomic dist
    rad_dict = {
        "RR": R_CIF_rad,
        "MM": M_CIF_rad,
        "XX": X_CIF_rad,
        "RM": R_CIF_rad + M_CIF_rad,
        "MX": M_CIF_rad + X_CIF_rad,
        "RX": R_CIF_rad + X_CIF_rad,
    }

    refined_rad_dict = {
        "RR": R_CIF_rad_refined,
        "MM": M_CIF_rad_refined,
        "XX": X_CIF_rad_refined,
        "RM": R_CIF_rad_refined + M_CIF_rad_refined,
        "MX": M_CIF_rad_refined + X_CIF_rad_refined,
        "RX": R_CIF_rad_refined + X_CIF_rad_refined,
    }
    shortest_homo_key = helper.get_shortest_homo_key(min_bond_dists, shortest_homoatomic_dist)
    shortest_hetero_key = helper.get_shortest_hetero_key(min_bond_dists, shortest_heteroatomic_dist)
    uni_features = helper.get_shortest_homo_heteroatomic_features(
        shortest_homoatomic_dist,
        shortest_homo_key,
        shortest_heteroatomic_dist,
        shortest_hetero_key,
        rad_dict,
        refined_rad_dict,
    )
    percent_diffs = [percent_diff_R, percent_diff_M, percent_diff_X]
    highest_refined_percent_diff = helper.get_highest_refined_percent_diff(percent_diffs)
    lowest_refined_percent_diff = helper.get_lowest_refined_percent_diff(percent_diffs)
    # Packing efficiency
    packing_efficiency = packing.compute_efficiency(
        cif,
        {
            R: R_CIF_rad_refined,
            M: M_CIF_rad_refined,
            X: X_CIF_rad_refined,
        },
    )
    data = {
        "Entry": cif.file_name_without_ext,
        "Formula": cif.formula,
        "Structure": cif.structure,
        "R": R,
        "M": M,
        "X": X,
        # 1. Shortest distances
        "INT_RR_dist": RR_dist,
        "INT_MM_dist": MM_dist,
        "INT_XX_dist": XX_dist,
        "INT_RM_dist": RM_dist,
        "INT_MX_dist": MX_dist,
        "INT_RX_dist": RX_dist,
        # 2. CIF radius
        "INT_Rsize": R_CIF_rad,
        "INT_Msize": M_CIF_rad,
        "INT_Xsize": X_CIF_rad,
        # 3. CIF radius ratio
        "INT_Rsize_by_Msize": Rsize_by_Msize,
        "INT_Msize_by_Xsize": Msize_by_Xsize,
        "INT_Rsize_by_Xsize": Rsize_by_Xsize,
        # 4. Radius from distance divided by CIF radius
        "INT_RR_dist_by_2_by_Rsize": RR_dist_by_2_by_Rsize,
        "INT_MM_dist_by_2_by_Msize": MM_dist_by_2_by_Msize,
        "INT_XX_dist_by_2_by_Xsize": XX_dist_by_2_by_Xsize,
        "INT_RM_dist_by_RMsizes": RM_dist_by_RM_sizes,
        "INT_MX_dist_by_MXsizes": MX_dist_by_MX_sizes,
        "INT_RX_dist_by_RXsizes": RX_dist_by_RX_sizes,
        # 5. Refined CIF radius
        "INT_Rsize_ref": R_CIF_rad_refined,
        "INT_Msize_ref": M_CIF_rad_refined,
        "INT_Xsize_ref": X_CIF_rad_refined,
        # 6. Difference between CIF radius and refined CIF radius
        "INT_percent_diff_R_by_100": percent_diff_R,
        "INT_percent_diff_M_by_100": percent_diff_M,
        "INT_percent_diff_X_by_100": percent_diff_X,
        # 7. Difference between interatomic distance and refined radius sum
        "INT_RR_dist_minus_ref_diff": RR_minus_ref_diff,
        "INT_MM_dist_minus_ref_diff": MM_minus_ref_diff,
        "INT_XX_dist_minus_ref_diff": XX_minus_ref_diff,
        "INT_RM_dist_minus_ref_diff": RM_minus_ref_diff,
        "INT_MX_dist_minus_ref_diff": MX_minus_ref_diff,
        "INT_RX_dist_minus_ref_diff": RX_minus_ref_diff,
        # 8. Objective value of the radius optimization
        "INT_R_factor": obj_value,
        # 9. Universal features
        "INT_UNI_shortest_homoatomic_dist": shortest_homoatomic_dist,
        "INT_UNI_shortest_heteroatomic_dist": shortest_heteroatomic_dist,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_atom_size"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": uni_features["shortest_heteroatomic_dist_by_sum_of_atom_sizes"],
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_refined_atom_sizes"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": uni_features["shortest_heteroatomic_dist_by_refined_atom_sizes"],
        "INT_UNI_highest_refined_percent_diff_abs": highest_refined_percent_diff,
        "INT_UNI_lowest_refined_percent_diff_abs": lowest_refined_percent_diff,
        "INT_UNI_refined_packing_efficiency": packing_efficiency,
    }
    uni_data = {
        "Entry": cif.file_name_without_ext,
        "Formula": cif.formula,
        "INT_UNI_shortest_homoatomic_dist": shortest_homoatomic_dist,
        "INT_UNI_shortest_heteroatomic_dist": shortest_heteroatomic_dist,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_atom_size"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": uni_features["shortest_heteroatomic_dist_by_sum_of_atom_sizes"],
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_refined_atom_sizes"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": uni_features["shortest_heteroatomic_dist_by_refined_atom_sizes"],
        "INT_UNI_highest_refined_percent_diff_abs": highest_refined_percent_diff,
        "INT_UNI_lowest_refined_percent_diff_abs": lowest_refined_percent_diff,
        "INT_UNI_refined_packing_efficiency": packing_efficiency,
    }
    return data, uni_data



from cifkit.utils import string_parser

from SAF.features.environment.util import _get_site_labels_per_element, count_first_second_min_dist, extract_best_labels


def compute_homoatomic_dist_by_site_shortest_dist(connections, A_best_label, B_best_label):
    """Compute the shortest homoatomic distances normalized by the
    shortest distance encountered at the best site for each element."""
    A = string_parser.get_atom_type_from_label(A_best_label)
    B = string_parser.get_atom_type_from_label(B_best_label)
    A_homoatomic_shortest_dist = float("inf")
    B_homoatomic_shortest_dist = float("inf")
    A_min_dist = float("inf")
    B_min_dist = float("inf")
    # Process each site
    for site_label, site_connections in connections.items():
        # Compute the shortest distance at this site
        current_min_dist = min([dist for _, dist, _, _ in site_connections])
        # Update the minimum distances for A and B if this is their best site
        if site_label == A_best_label:
            A_min_dist = current_min_dist
            # Find the shortest homoatomic distance for A
            A_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(A)])

        if site_label == B_best_label:
            B_min_dist = current_min_dist
            # Find the shortest homoatomic distance for B
            B_homoatomic_shortest_dist = min([dist for label, dist, _, _ in site_connections if label.startswith(B)])
    # Normalize the shortest distances by the shortest distances at the best sites
    A_homoatomic_dist_by_shortest_dist = A_homoatomic_shortest_dist / A_min_dist
    B_homoatomic_dist_by_shortest_dist = B_homoatomic_shortest_dist / B_min_dist

    return (
        A_homoatomic_dist_by_shortest_dist,
        B_homoatomic_dist_by_shortest_dist,
    )


def compute_avg_homoatomic_dist_by_site_shortest_dist(connections, A, B):
    """Compute the average of the homoatomic distances normalized by the
    shortest distance encountered at each site for two distinct element
    types within a given set of connections."""
    # Shortest distance per site label
    element_to_site_labels = _get_site_labels_per_element(connections)
    A_total_homoatomic_dist_by_shortest_dist = 0.0
    B_total_homoatomic_dist_by_shortest_dist = 0.0

    for site_label, site_connections in connections.items():
        min_dist_per_site = site_connections[0][1]
        for connection in site_connections:
            other_label, dist, _, _ = connection
            if site_label.startswith(A) and other_label.startswith(A):
                A_total_homoatomic_dist_by_shortest_dist += dist / min_dist_per_site
                break
            if site_label.startswith(B) and other_label.startswith(B):
                B_total_homoatomic_dist_by_shortest_dist += dist / min_dist_per_site
                break

    # Average of (A-A distance / shortest distance for each site label)
    # Average of (B-B distance / (shortest distance for each site label)
    A_avg_homoatomic_dist_by_shortest_dist = A_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[A])
    B_avg_homoatomic_dist_by_shortest_dist = B_total_homoatomic_dist_by_shortest_dist / len(element_to_site_labels[B])
    return (
        A_avg_homoatomic_dist_by_shortest_dist,
        B_avg_homoatomic_dist_by_shortest_dist,
    )


def get_A_and_B_count_in_best_label_per_element(connections, A, B):
    """Compute the occurrences of element types A and B at the shortest
    distances within their respective best labeled site for each
    element.

    The best site label is determined with the label with the shortest
    distance pair.
    """
    A_count_at_A_shortest_dist = 0
    A_count_at_B_shortest_dist = 0
    B_count_at_A_shortest_dist = 0
    B_count_at_B_shortest_dist = 0
    dist_count_per_label = count_first_second_min_dist(connections)
    best_label_dict = extract_best_labels(dist_count_per_label)
    best_A_site_label = best_label_dict[A]["best_label"]
    best_A_shortest_dist = best_label_dict[A]["best_label_details"]["shortest_dist"]
    best_B_site_label = best_label_dict[B]["best_label"]
    best_B_shortest_dist = best_label_dict[B]["best_label_details"]["shortest_dist"]
    for connection in connections[best_A_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_A_shortest_dist:
            if other_label.startswith(A):
                A_count_at_A_shortest_dist += 1
            if other_label.startswith(B):
                B_count_at_A_shortest_dist += 1
    for connection in connections[best_B_site_label]:
        other_label, dist, _, _ = connection
        if dist == best_B_shortest_dist:
            if other_label.startswith(A):
                A_count_at_B_shortest_dist += 1
            if other_label.startswith(B):
                B_count_at_B_shortest_dist += 1
    return (
        A_count_at_A_shortest_dist,
        A_count_at_B_shortest_dist,
        B_count_at_A_shortest_dist,
        B_count_at_B_shortest_dist,
    )


def get_avg_A_and_B_count_in_per_element(connections, A, B):
    """Compute the occurrences of element types A and B at the shortest
    distances within their respective site for all site labels."""
    A_total_count_at_A_shortest_dist = 0
    A_total_count_at_B_shortest_dist = 0
    B_total_count_at_A_shortest_dist = 0
    B_total_count_at_B_shortest_dist = 0

    dist_count_per_label = count_first_second_min_dist(connections)
    # Use the best label dict to parse the number of site labels for A and B
    best_labels = extract_best_labels(dist_count_per_label)
    A_site_label_count = best_labels[A]["label_count"]
    B_site_label_count = best_labels[B]["label_count"]

    for site_label, connection_data in connections.items():
        # Get the shortest distance for the site label
        parsed_element = string_parser.get_atom_type_from_label(site_label)
        shortest_dist_per_site_label = dist_count_per_label[site_label]["shortest_dist"]

        # Loop through each connection
        for connection in connection_data:
            other_label, dist, _, _ = connection
            if parsed_element == A:
                if dist == shortest_dist_per_site_label:
                    if other_label.startswith(A):
                        A_total_count_at_A_shortest_dist += 1
                    if other_label.startswith(B):
                        B_total_count_at_A_shortest_dist += 1

            if parsed_element == B:
                if dist == shortest_dist_per_site_label:
                    if other_label.startswith(A):
                        A_total_count_at_B_shortest_dist += 1
                    if other_label.startswith(B):
                        B_total_count_at_B_shortest_dist += 1
    A_avg_count_at_A_shortest_dist = A_total_count_at_A_shortest_dist / A_site_label_count
    A_avg_count_at_B_shortest_dist = A_total_count_at_B_shortest_dist / B_site_label_count
    B_avg_count_at_A_shortest_dist = B_total_count_at_A_shortest_dist / A_site_label_count
    B_avg_count_at_B_shortest_dist = B_total_count_at_B_shortest_dist / B_site_label_count
    return (
        A_avg_count_at_A_shortest_dist,
        A_avg_count_at_B_shortest_dist,
        B_avg_count_at_A_shortest_dist,
        B_avg_count_at_B_shortest_dist,
    )



from cifkit.utils.string_parser import strip_numbers_and_symbols


def _parse_site_data(loop_values):
    # Initialize a dictionary to store element information
    data = {}
    # Get the number of atoms
    num_site_labels = len(loop_values[0])
    # Loop over all atoms
    for i in range(num_site_labels):
        # Get atomic info
        site_label = loop_values[0][i]
        element = strip_numbers_and_symbols(loop_values[1][i])
        multiplicity = int(loop_values[2][i])
        if element not in data:
            data[element] = {
                "sites": 0,
                "multiplicity": 0,
                "lowest_wyckoff_multiplicity": multiplicity,
                "lowest_wyckoff_element": site_label,
            }
        # Update the atom information in the dictionary
        data[element]["sites"] += 1
        data[element]["multiplicity"] += multiplicity
        # Update the element with the lowest Wyckoff multiplicity
        if multiplicity < data[element]["multiplicity"]:
            data[element]["lowest_wyckoff_multiplicity"] = multiplicity
            data[element]["lowest_wyckoff_element"] = site_label
    return data


def get_binary_site_info(loop_values, A, B):
    atomic_env = _parse_site_data(loop_values)
    A_env, B_env = None, None
    # Check if the desired elements are present
    if A in atomic_env:
        A_env = atomic_env[A]
    if B in atomic_env:
        B_env = atomic_env[B]
    return A_env, B_env


def get_ternary_site_info(loop_values, R, M, X):
    atomic_env = _parse_site_data(loop_values)
    R_env, M_env, X_env = None, None, None
    # Check if the desired elements are present
    if R in atomic_env:
        R_env = atomic_env[R]
    if M in atomic_env:
        M_env = atomic_env[M]
    if X in atomic_env:
        X_env = atomic_env[X]
    return R_env, M_env, X_env


def get_quaternary_site_info(loop_values, A, B, C, D):
    atomic_env = _parse_site_data(loop_values)
    A_env, B_env, C_env, D_env = None, None, None, None
    # Check if the desired elements are present
    if A in atomic_env:
        A_env = atomic_env[A]
    if B in atomic_env:
        B_env = atomic_env[B]
    if C in atomic_env:
        C_env = atomic_env[C]
    if D in atomic_env:
        D_env = atomic_env[D]
    return A_env, B_env, C_env, D_env
